#!/usr/bin/env python
"""
Calculate output results dictionary.
"""

from collections import OrderedDict as odict

import numpy as np
import yaml
import numpy.lib.recfunctions as recfuncs

import astropy.coordinates
from astropy.coordinates import SkyCoord
import astropy.units as u

import ugali.analysis.source
import ugali.analysis.loglike

import ugali.utils.stats
from ugali.utils.stats import Samples
from ugali.utils.projector import dist2mod,mod2dist
from ugali.utils.projector import cel2gal,gal2cel,gal2cel_angle
from ugali.utils.projector import ang2const, ang2iau
from ugali.utils.config import Config
from ugali.utils.logger import logger

_alpha = 0.32

class Results(object):
    """
    Calculate results from a MCMC chain.
    """
    def __init__(self, config, loglike, samples=None):
        self.config = Config(config)
        self.alpha = self.config['results'].get('alpha',_alpha)
        self.nwalkers = self.config['mcmc'].get('nwalkers',100)
        self.nburn = self.config['results'].get('nburn',10)
        self.coordsys = self.config['coords']['coordsys'].lower()
        
        self.loglike = loglike
        self.source = self.loglike.source
        self.params = list(self.source.get_free_params().keys())
        self.samples = samples

    def load_samples(self,filename):
        samples = Samples(filename)
        self.samples = samples.supplement(coordsys=self.coordsys)


    def get_mle(self):
        mle = self.source.get_params()
        # FIXME: For composite isochrones
        if 'age' not in mle:
            mle['age'] = np.average(self.source.isochrone.age)
        if 'metallicity' not in mle:
            mle['metallicity'] = np.average(self.source.isochrone.metallicity)
            
        return mle

    def estimate(self,param,burn=None,clip=10.0,alpha=_alpha):
        """ Estimate parameter value and uncertainties 

        Parameters
        ----------
        param : parameter of interest
        burn  : number of burn in samples
        clip  : sigma clipping
        alpha : confidence interval

        Returns
        -------
        [mle, [lo,hi]] : value and interval
        """
        # FIXME: Need to add age and metallicity to composite isochrone params (currently properties)
        if alpha is None: alpha = self.alpha
        if param not in list(self.samples.names) + list(self.source.params) + ['age','metallicity']:
            msg = 'Unrecognized parameter: %s'%param
            raise KeyError(msg)

        # If the parameter is in the samples
        if param in self.samples.names:
            if param.startswith('position_angle'):
                return self.estimate_position_angle(param,burn=burn,
                                                    clip=clip,alpha=alpha)

            return self.samples.peak_interval(param,burn=burn,clip=clip,alpha=alpha)
 
        mle = self.get_mle()
        errors = [np.nan,np.nan] 

        # Set default value to the MLE value
        if param in self.source.params:
            err = self.source.params[param].errors
            if err is not None: errors = err

        # For age and metallicity from composite isochrone
        return [float(mle[param]),errors]

        ### if (param not in self.params) or (param not in : 
        ###     return [float(mle[param]),errors]
        ###  
        ### if param not in self.samples.names: 
        ###     return [float(mle[param]),errors]
        ###  
        ### msg = "Unrecognized parameter: %s"%param
        ### raise KeyError(msg)
 
    def estimate_params(self,burn=None,clip=10.0,alpha=None):
        """ Estimate all source parameters """
        if alpha is None: alpha = self.alpha
        mle = self.get_mle()
        out = odict()
        for param in mle.keys():
            out[param] = self.estimate(param,burn=burn,clip=clip,alpha=alpha)
        return out
 
    def estimate_position_angle(self,param='position_angle',burn=None,clip=10.0,alpha=None):
        """ Estimate the position angle from the posterior dealing
        with periodicity.
        """
        if alpha is None: alpha = self.alpha
        # Transform so peak in the middle of the distribution
        pa = self.samples.get(param,burn=burn,clip=clip)
        peak = ugali.utils.stats.kde_peak(pa)
        shift = 180.*((pa+90-peak)>180)
        pa -= shift
        # Get the kde interval
        ret = ugali.utils.stats.peak_interval(pa,alpha)
        if ret[0] < 0: 
            ret[0] += 180.; ret[1][0] += 180.; ret[1][1] += 180.;
        return ret
 
    def bayes_factor(self,param,burn=None,clip=10.0,bins=50):
        # CAREFUL: Assumes a flat prior...
        try: 
            data = self.samples.get(param,burn=burn,clip=clip)
        except ValueError as msg:
            logger.warning(msg)
            return ugali.utils.stats.interval(np.nan)
 
        bmin,bmax = self.source.params[param].bounds
        bins = np.linspace(bmin,bmax,bins)
        n,b = np.histogram(data,bins=bins,normed=True)
        prior = 1.0/(bmax-bmin)
        posterior = n[0]
        # Excluding the null hypothesis
        bf = prior/posterior
        return ugali.utils.stats.interval(bf)
 
    def get_results(self,**kwargs):
        kwargs.setdefault('alpha',self.alpha)
        kwargs.setdefault('burn',self.nburn*self.nwalkers)

        # Calculate best-fit parameters from MCMC chain
        logger.debug('Estimating parameters...')
        estimate = self.estimate_params(**kwargs)
        params = {k:v[0] for k,v in estimate.items()}
        results = dict(estimate)

        # Extra parameters from the MCMC chain
        logger.debug('Estimating auxiliary parameters...')
        logger.debug("alpha = %.2f"%kwargs['alpha'])
        results['alpha'] = kwargs['alpha']
        try: 
            results['ra']  = self.estimate('ra',**kwargs)
            results['dec'] = self.estimate('dec',**kwargs)
            results['glon'] = self.estimate('glon',**kwargs)
            results['glat'] = self.estimate('glat',**kwargs)
        except KeyError:
            logger.warn("Didn't find 'ra' or 'dec' in Samples...")
            if self.coordsys == 'gal':
                results['glon'] = results['lon']
                results['glat'] = results['lat']
                ra,dec = gal2cel(results['lon'][0],results['lat'][0])
                results['ra'] = ugali.utils.stats.interval(ra)
                results['dec'] = ugali.utils.stats.interval(dec)
            else:
                results['ra'] = results['lon']
                results['dec'] = results['lat']
                glon,glat = cel2gal(results['lon'][0],results['lat'][0])
                results['glon'] = ugali.utils.stats.interval(glon)
                results['glat'] = ugali.utils.stats.interval(glat)

        lon,lat   = results['lon'][0],results['lat'][0]
        ra,dec    = results['ra'][0],results['dec'][0]
        glon,glat = results['glon'][0],results['glat'][0]
        results.update(gal=[float(glon),float(glat)])
        results.update(cel=[float(ra),float(dec)])

        try:
            results['position_angle_cel']  = self.estimate('position_angle_cel',**kwargs)
        except KeyError:
            results['position_angle_cel'] = ugali.utils.stats.interval(np.nan)

        # Update the loglike to the best-fit parameters from the chain
        logger.debug('Calculating TS...')
        ts = 2*self.loglike.value(**params)
        results['ts'] = ugali.utils.stats.interval(ts,np.nan,np.nan)
 
        # Celestial position angle
        # Break ambiguity in direction with '% 180.'
        pa,pa_err = results['position_angle']
        pa_cel = gal2cel_angle(lon,lat,pa) % 180.
        pa_cel_err = np.array(pa_err) - pa + pa_cel
        results['position_angle_cel'] = ugali.utils.stats.interval(pa_cel,pa_cel_err[0],pa_cel_err[1])
        
        mod,mod_err = estimate['distance_modulus']
        dist = mod2dist(mod)
        dist_lo,dist_hi = [mod2dist(mod_err[0]),mod2dist(mod_err[1])]
        results['distance'] = ugali.utils.stats.interval(dist,dist_lo,dist_hi)
        dist,dist_err = results['distance']
 
        ext,ext_err = estimate['extension']
        ext_sigma = np.nan_to_num(np.array(ext_err) - ext)
        results['extension_arcmin'] = ugali.utils.stats.interval(60*ext,60*ext_err[0],60*ext_err[1])
 
        # Radially symmetric extension (correct for ellipticity).
        ell,ell_err = estimate['ellipticity']
        rext,rext_err = ext*np.sqrt(1-ell),np.array(ext_err)*np.sqrt(1-ell)
        rext_sigma = np.nan_to_num(np.array(rext_err) - rext)
        results['extension_radial'] = ugali.utils.stats.interval(rext,rext_err[0],rext_err[1])
        results['extension_radial_arcmin'] = ugali.utils.stats.interval(60*rext,60*rext_err[0],60*rext_err[1])
 
        # Bayes factor for ellipticity
        results['ellipticity_bayes_factor'] = self.bayes_factor('ellipticity',burn=kwargs['burn'])
 
        # Physical Size (should do this with the posteriors)
        # Radially symmetric
        dist_sigma = np.nan_to_num(np.array(dist_err) - dist)

        size = np.arctan(np.radians(ext)) * dist
        size_sigma = size * np.sqrt((ext_sigma/ext)**2 + (dist_sigma/dist)**2)
        size_err = [size-size_sigma[0],size+size_sigma[1]]
        results['physical_size'] = ugali.utils.stats.interval(size,size_err[0],size_err[1])

        rsize = np.arctan(np.radians(rext)) * dist
        rsize_sigma = rsize * np.sqrt((rext_sigma/rext)**2 + (dist_sigma/dist)**2)
        rsize_err = [rsize-rsize_sigma[0],rsize+rsize_sigma[1]]
        results['physical_size_radial'] = ugali.utils.stats.interval(rsize,rsize_err[0],rsize_err[1])
 
        # Richness
        rich,rich_err = estimate['richness']
 
        # Number of observed stars (sum of p-values)
        nobs = self.loglike.p.sum()
        nobs_lo,nobs_hi = nobs + np.sqrt(nobs)*np.array([-1,1])
        results['nobs'] = ugali.utils.stats.interval(nobs,nobs_lo,nobs_hi)
 
        # Number of predicted stars (pixelization effects?)
        npred = self.loglike.f*rich
        npred_lo,npred_hi = rich_err[0]*self.loglike.f,rich_err[1]*self.loglike.f
        results['npred'] = ugali.utils.stats.interval(npred,npred_lo,npred_hi)
        
        # Careful, depends on the isochrone...
        stellar_mass = self.source.stellar_mass()
        mass = rich*stellar_mass
        mass_lo,mass_hi = rich_err[0]*stellar_mass,rich_err[1]*stellar_mass
        results['mass'] = ugali.utils.stats.interval(mass,mass_lo,mass_hi)
 
        stellar_luminosity = self.source.stellar_luminosity()
        lum = rich*stellar_luminosity
        lum_lo,lum_hi = rich_err[0]*stellar_luminosity,rich_err[1]*stellar_luminosity
        results['luminosity'] = ugali.utils.stats.interval(lum,lum_lo,lum_hi)

        # Absolute magnitude only calculated for DES isochrones with g,r 
        try:
            Mv = self.source.absolute_magnitude(rich)
            Mv_lo = self.source.absolute_magnitude(rich_err[0])
            Mv_hi = self.source.absolute_magnitude(rich_err[1])
            results['Mv'] = ugali.utils.stats.interval(Mv,Mv_lo,Mv_hi)
        except ValueError as e:
            logger.warning("Skipping absolute magnitude")
            logger.warn(str(e))
            Mv = np.nan
            results['Mv'] = Mv

        mu = surfaceBrightness(Mv, rsize, dist) ##updated from size-->rsize
        results['surface_brightness'] = ugali.utils.stats.interval(mu,np.nan,np.nan)

        # ADW: WARNING this is very fragile.
        # Also, this is not quite right, should cut on the CMD available space
        kwargs = dict(richness=rich,mag_bright=16., mag_faint=23.,
                      n_trials=5000,alpha=kwargs['alpha'], seed=0)
        martin = self.config['results'].get('martin')
        if martin:
            logger.info("Calculating Martin magnitude...")
            if martin > 1: kwargs['n_trials'] = martin
            Mv_martin = self.source.isochrone.absolute_magnitude_martin(**kwargs)
            results['Mv_martin'] = Mv_martin

            mu_martin = surfaceBrightness(Mv_martin, rsize, dist) ##updated from size-->rsize
            results['surface_brightness_martin'] = ugali.utils.stats.interval(mu_martin,np.nan,np.nan)
        else:
            logger.warning("Skipping Martin magnitude")
            results['Mv_martin'] = np.nan
            results['surface_brightness_martin'] = np.nan

        try: 
            results['constellation'] = ang2const(lon,lat,self.coordsys)[1]
        except:
            pass
        results['iau'] = ugali.utils.projector.ang2iau(lon,lat,self.coordsys)
 
        coord = SkyCoord(ra*u.deg,dec*u.deg,distance=dist*u.kpc)
        results['ra_sex'] = str(coord.ra.to_string())
        results['dec_sex'] = str(coord.dec.to_string())
 
        # Calculate some separations from GC, LMC, SMC
        #NED coordinates with de Grisj distance
        LMC = SkyCoord(80.8939*u.deg,-69.7561*u.deg,distance=49.89*u.kpc)
        #NED coordinates with de Grisj distance
        SMC = SkyCoord(13.1866*u.deg,-72.8286*u.deg,distance=61.94*u.kpc)
        # GC from astropy?
        GC = SkyCoord(266.4168262*u.deg,-29.0077969*u.deg,distance=8.0*u.kpc)
         
        results['d_gc'] = coord.separation_3d(GC).value
        results['d_lmc'] = coord.separation_3d(LMC).value
        results['d_smc'] = coord.separation_3d(SMC).value

        try:
            results['feh'] = float(self.source.isochrone.feh)
        except:
            results['feh'] = np.nan
        
        output = dict()
        output['params'] = params
        output['results'] = results
        return output

    def write(self,filename):
        if self.samples is not None:
            results = dict(self.get_results())
            params  = dict(params=results.pop('params'))
        else:
            results = dict(results=dict())
            params = dict(params=dict())
        source = dict(source=self.source.todict())
 
        out = open(filename,'w')
        out.write(yaml.dump(params,default_flow_style=False))
        out.write(yaml.dump(results))
        out.write(yaml.dump(source))
        out.close()

def surfaceBrightness(abs_mag, r_physical, distance):
    """Compute the average surface brightness [mag/arcsec^2] within the
    azimuthally averaged physical half-light radius.

    The azimuthally averaged physical half-light radius is used
    because the surface brightness is calculated as a function of
    area. The area of an ellipse is:

    A = pi * a*b = pi * a**2 * (1-e) = pi * r**2

    The factor 2 in the c_v equation below accounts for the fact that
    we are taking half the luminosity within the half-light
    radius. The 3600.**2 is conversion from deg^2 to arcsec^2

    c_v = 2.5 * np.log10(2.) + 2.5 * np.log10(np.pi * 3600.**2) = 19.78

    TODO: Distance could be optional

    Parameters
    ----------
    abs_mag    : absolute V-band magnitude [mag]
    r_physical : azimuthally averaged physical half-light radius [kpc]
    distance   : heliocentric distance [kpc]

    Returns
    -------
    mu         : surface brightness [mag/arcsec^2]
    """
    c_v = 19.78 # conversion to mag/arcsec^2
    r_angle = np.degrees(np.arctan(r_physical / distance))
    return abs_mag + dist2mod(distance) + c_v + 2.5 * np.log10(r_angle**2)

def surfaceBrightness2(app_mag, r_half_arcmin):
    """Compute the average surface brightness [mag/arcsec^2] within the
    azimuthally averaged angular half-light radius.

    The azimuthally averaged half-light radius is used
    because the surface brightness is calculated as a function of
    area. The area of an ellipse is:

    A = pi * a*b = pi * a**2 * (1-e) = pi * r**2

    The factor 2.5*log10(2) accounts for the fact that half the
    luminosity is within the half-light radius.

    Parameters
    ----------
    app_mag : apparent V-band magnitude [mag]
    r_half  : azimuthally averaged half-light radius [arcmin]

    Returns
    -------
    mu      : surface brightness [mag/arcsec^2]

    """
    r_half_arcsec = r_half_arcmin * 60 # arcmin to arcsec
    mu = app_mag + 2.5*np.log10(2.) + 2.5*np.log10(np.pi * r_half_arcsec**2)
    return mu


def createResults(config,srcfile,section='source',samples=None):
    """ Create an MCMC instance """
    source = ugali.analysis.source.Source()
    source.load(srcfile,section=section)
    loglike = ugali.analysis.loglike.createLoglike(config,source)

    results = Results(config,loglike,samples)

    if samples is not None:
        results.load_samples(samples)

    return results

def write_results(filename,config,srcfile,samples):
    """ Package everything nicely """ 
    results = createResults(config,srcfile,samples=samples)
    results.write(filename)


if __name__ == "__main__":
    import ugali.utils.parser
    parser = ugali.utils.parser.Parser(description=__doc__)
    parser.add_config()
    parser.add_verbose()
    parser.add_argument('--srcmdl',required=True,
                        help='Source model file')
    parser.add_argument('--section',default='source',
                        help='Section of source file')
    parser.add_argument('--samples',required=True,
                        help='Posterior samples file')
    parser.add_argument('outfile',default=None,
                        help="Output file name")
                        
    args = parser.parse_args()

    #write_results(args.outfile,args.config,args.srcmdl,args.samples)
    results = createResults(args.config,args.srcmdl,samples=args.samples)
    results.write(args.outfile)




#!/usr/bin/env python
"""
Download Padova isochrones from:
http://stev.oapd.inaf.it/cgi-bin/cmd

Adapted from ezpadova by Morgan Fouesneau:
https://github.com/mfouesneau/ezpadova

"""
import os
try:
    from urllib.parse import urlencode
    from urllib.request import urlopen
except ImportError:
    from urllib import urlencode
    from urllib2 import urlopen
import re
import subprocess
from multiprocessing import Pool
from collections import OrderedDict as odict
import copy

import numpy as np
from ugali.utils.logger import logger
from ugali.utils.shell import mkdir
from ugali.analysis.isochrone import Padova

# survey system
photsys_dict = odict([
        ('des' ,'tab_mag_odfnew/tab_mag_decam.dat'),
        ('sdss','tab_mag_odfnew/tab_mag_sloan.dat'),
        ('ps1' ,'tab_mag_odfnew/tab_mag_panstarrs1.dat'),
        ('lsst','tab_mag_odfnew/tab_mag_lsst.dat'),
])

photname_dict = odict([
        ('des' ,'DECAM'),
        ('sdss','SDSS'),
        ('ps1' ,'Pan-STARRS1'),
        ('lsst','LSST'),
])

# Commented options may need to be restored for older version/isochrones.
# The parameters were tracked down by:
# Chrome -> View -> Developer -> Developer Tools
# Network -> Headers -> Request Payload

defaults_cmd=  {#'binary_frac': 0.3,
                #'binary_kind': 1,
                #'binary_mrinf': 0.7,
                #'binary_mrsup': 1,
                'cmd_version': 2.7,
                'dust_source': 'nodust',
                'dust_sourceC': 'nodustC',
                'dust_sourceM': 'nodustM',
                'eta_reimers': 0.2,
                #'extinction_av': 0,
                #'icm_lim': 4,
                'imf_file': 'tab_imf/imf_chabrier_lognormal.dat',
                'isoc_age': 1e9,
                'isoc_age0': 12.7e9,
                'isoc_dlage': 0.05,
                'isoc_dz': 0.0001,
                'isoc_kind': 'parsec_CAF09_v1.2S',
                'isoc_lage0': 6.602,   #Minimum allowed age                 
                'isoc_lage1': 10.1303, #Maximum allowed age                 
                'isoc_val': 0,               
                'isoc_z0': 0.0001,     #Minimum allowed metallicity         
                'isoc_z1': 0.03,       #Maximum allowed metallicity   
                'isoc_zeta': 0.0002,
                'isoc_zeta0': 0.0002,
                'kind_cspecmag': 'aringer09',
                'kind_dust': 0,
                'kind_interp': 1,
                'kind_mag': 2,
                'kind_postagb': -1,
                'kind_pulsecycle': 0,
                #'kind_tpagb': 0,
                #'lf_deltamag': 0.2,
                #'lf_maginf': 20,
                #'lf_magsup': -20,
                #'mag_lim': 26,
                #'mag_res': 0.1,
                'output_evstage': 1,
                'output_gzip': 0,
                'output_kind': 0,
                'photsys_file': photsys_dict['des'],
                #'photsys_version': 'yang',
                'submit_form': 'Submit'}

defaults_27 = dict(defaults_cmd,cmd_version='2.7')
defaults_28 = dict(defaults_cmd,cmd_version='2.8')
defaults_29 = dict(defaults_cmd,cmd_version='2.9')
defaults_30 = dict(defaults_cmd,cmd_version='3.0')


class Download(object):

    isochrone = None
    
    def __init__(self,survey='des',**kwargs):
        self.survey=survey.lower()

    def create_grid(self,abins,zbins):
        arange = np.linspace(abins[0],abins[1],abins[2]+1)
        zrange = np.logspace(np.log10(zbins[0]),np.log10(zbins[1]),zbins[2]+1)
        aa,zz = np.meshgrid(arange,zrange)
        return aa.flatten(),zz.flatten()

    def print_info(self,age,metallicity):
        params = dict(age=age,z=metallicity)
        params['name'] = self.__class__.__name__
        params['survey'] = self.survey
        params['feh'] = self.isochrone.z2feh(metallicity)
        msg = 'Downloading: %(name)s (survey=%(survey)s, age=%(age).1fGyr, Z=%(z).5f, Fe/H=%(feh).3f)'%params
        logger.info(msg)
        return msg

    def query_server(self,outfile,age,metallicity):
        msg = "'query_server' not implemented by base class."
        logger.error(msg)
        raise RuntimeError(msg)

    @classmethod
    def verify(cls,filename,survey,age,metallicity):
        msg = "'verify' not implemented by base class."
        logger.error(msg)
        raise RuntimeError(msg)

    def download(self,age,metallicity,outdir=None,force=False):
        """
        Check valid parameter range and download isochrones from:
        http://stev.oapd.inaf.it/cgi-bin/cmd
        """
        if outdir is None: outdir = './'
        basename = self.isochrone.params2filename(age,metallicity)
        outfile = os.path.join(outdir,basename)
            
        if os.path.exists(outfile) and not force:
            try:
                self.verify(outfile,self.survey,age,metallicity)
                logger.info("Found %s; skipping..."%(outfile))
                return
            except Exception as e:
                msg = "Overwriting corrupted %s..."%(outfile)
                logger.warn(msg)
                #os.remove(outfile)
                
        mkdir(outdir)

        self.print_info(age,metallicity)

        try:
            self.query_server(outfile,age,metallicity)
        except Exception as e:
            logger.debug(str(e))
            raise RuntimeError('Bad server response')

        if not os.path.exists(outfile):
            raise RuntimeError('Download failed')

        try:
            self.verify(outfile,self.survey,age,metallicity)
        except Exception as e:
            msg = "Output file is corrupted."
            logger.error(msg)
            #os.remove(outfile)
            raise(e)

        return outfile

class Padova(Download):
    defaults = copy.deepcopy(defaults_27)
    isochrone = Padova
    
    abins = np.arange(1.0, 13.5 + 0.1, 0.1)
    zbins = np.arange(1e-4,1e-3 + 1e-5,1e-5)

    def query_server(self,outfile,age,metallicity):
        epsilon = 1e-4
        lage = np.log10(age*1e9)
        lage_min,lage_max = self.defaults['isoc_lage0'],self.defaults['isoc_lage1']
        if not (lage_min-epsilon < lage <lage_max+epsilon):
            msg = 'Age outside of valid range: %g [%g < log(age) < %g]'%(lage,lage_min,lage_max)
            raise RuntimeError(msg)

        z_min,z_max = self.defaults['isoc_z0'],self.defaults['isoc_z1']
        if not (z_min <= metallicity <= z_max):
            msg = 'Metallicity outside of valid range: %g [%g < z < %g]'%(metallicity,z_min,z_max)
            raise RuntimeError(msg)

        d = dict(self.defaults)
        d['photsys_file'] = photsys_dict[self.survey]
        d['isoc_age']     = age * 1e9
        d['isoc_zeta']    = metallicity

        server = 'http://stev.oapd.inaf.it'
        url = server + '/cgi-bin/cmd_%s'%d['cmd_version']
        logger.debug("Accessing %s..."%url)

        q = urlencode(d)
        logger.debug(url+'?'+q)
        c = str(urlopen(url, str.encode(q)).read())
        aa = re.compile('output\d+')
        fname = aa.findall(c)
        
        if len(fname) == 0:
            msg = "Output filename not found"
            raise RuntimeError(msg)

        out = '{0}/tmp/{1}.dat'.format(server, fname[0])
        cmd = 'wget %s -O %s'%(out,outfile)
        logger.debug(cmd)
        stdout = subprocess.check_output(cmd,shell=True,stderr=subprocess.STDOUT)
        logger.debug(stdout)

        return outfile

    def verify(cls, filename, survey, age, metallicity):
        age = age*1e9
        nlines=14
        with open(filename,'r') as f:
            lines = [f.readline() for i in range(nlines)]
            if len(lines) < nlines:
                msg = "Incorrect file size"
                raise Exception(msg)
                
            try:
                s = lines[2].split()[-2]
                assert photname_dict[survey][:4] in s
            except:
                msg = "Incorrect survey:\n"+lines[2]
                raise Exception(msg)

            try:
                z = lines[5].split()[2]
                assert np.allclose(metallicity,float(z),atol=1e-3)
            except:
                msg = "Metallicity does not match:\n"+lines[5]
                raise Exception(msg)

            try:
                a = lines[13].split()[1]
                assert np.allclose(age,float(a),atol=1e-5)
            except:
                msg = "Age does not match:\n"+lines[13]
                raise Exception(msg)


    @classmethod
    def verify(cls, filename, survey, age, metallicity):
        age = age*1e9
        nlines=15
        with open(filename,'r') as f:
            lines = [f.readline() for i in range(nlines)]
            if len(lines) < nlines:
                msg = "Incorrect file size"
                raise Exception(msg)

            for i,l in enumerate(lines):
                if l.startswith('# Photometric system:'): break
            else:
                msg = "Incorrect file header"
                raise Exception(msg)

            try:
                s = lines[i].split()[3]
                assert photname_dict[survey] == s
            except:
                msg = "Incorrect survey:\n"+lines[i]
                raise Exception(msg)

            try:
                z = lines[-1].split()[0]
                assert np.allclose(metallicity,float(z),atol=1e-5)
            except:
                msg = "Metallicity does not match:\n"+lines[-1]
                raise Exception(msg)

            try:
                a = lines[-1].split()[1]
                # Need to deal with age or log-age
                assert (np.allclose(age,float(a),atol=1e-2) or
                        np.allclose(np.log10(age),float(a),atol=1e-2))
            except:
                msg = "Age does not match:\n"+lines[-1]
                raise Exception(msg)

class Girardi2002(Padova):
    defaults = dict(defaults_27)
    defaults['isoc_kind'] = 'gi2000'

class Marigo2008(Padova):
    defaults = dict(defaults_27)
    defaults['isoc_kind'] = 'ma08'

class Girardi2010a(Padova):
    defaults = dict(defaults_27)
    defaults['isoc_kind'] = 'gi10a'

class Girardi2010b(Padova):
    defaults = dict(defaults_27)
    defaults['isoc_kind'] = 'gi10b'

class Bressan2012(Padova):
    defaults = dict(defaults_27)
    defaults['isoc_kind'] = 'parsec_CAF09_v1.2S'

class Marigo2017(Padova):
    defaults = dict(defaults_30)
    defaults['isoc_kind'] = 'parsec_CAF09_v1.2S_NOV13'

def factory(name, **kwargs):
    from ugali.utils.factory import factory
    return factory(name, module=__name__, **kwargs)

if __name__ == "__main__":
    import ugali.utils.parser
    description = "Download isochrones"
    parser = ugali.utils.parser.Parser(description=description)
    parser.add_verbose()
    parser.add_force()
    parser.add_argument('-a','--age',default=None,type=float,action='append')
    parser.add_argument('-z','--metallicity',default=None,type=float,action='append')
    parser.add_argument('-k','--kind',default='Bressan2012')
    parser.add_argument('-s','--survey',default='des')
    parser.add_argument('-o','--outdir',default=None)
    parser.add_argument('-n','--njobs',default=10,type=int)
    args = parser.parse_args()

    if args.verbose:
        try:                                                                                            
            from http.client import HTTPConnection
        except ImportError:
            from httplib import HTTPConnection
        HTTPConnection.debuglevel = 1

    if args.outdir is None:
        args.outdir = os.path.join(args.survey.lower(),args.kind.lower())
    logger.info("Writing to output directory: %s"%args.outdir)

    p = factory(args.kind,survey=args.survey)

    # Defaults
    abins = [args.age] if args.age else p.abins
    zbins = [args.metallicity] if args.metallicity else p.zbins
    grid = [g.flatten() for g in np.meshgrid(abins,zbins)]
    logger.info("Ages:\n  %s"%np.unique(grid[0]))
    logger.info("Metallicities:\n  %s"%np.unique(grid[1]))

    def run(args):
        try:
            p.download(*args)
        except Exception as e:
            logger.warn(str(e))
            logger.error("Download failed.")

    arguments = [(a,z,args.outdir,args.force) for a,z in zip(*grid)]
    if args.njobs > 1:
        pool = Pool(processes=args.njobs, maxtasksperchild=100)
        results = pool.map(run,arguments)
    else:
        results = list(map(run,arguments))
    



#!/usr/bin/env python

#This is only for backwards compatibility
msg = "'import ugali.analysis.isochrone' is deprecated. "
msg += "Use 'import ugali.isochrone' instead."
DeprecationWarning(msg)

from ugali.isochrone import *



"""
Classes and functions that handle masks (i.e., exposure depth). 

Classes
    Mask

Functions
    someFunction
"""

import os
import numpy as np
import healpy as hp
import scipy.signal

import ugali.utils.binning
import ugali.utils.skymap

import ugali.observation.roi
from ugali.utils import healpix

from ugali.utils.logger import logger
from ugali.utils.healpix import ang2pix
from ugali.utils.config import Config
from ugali.utils.constants import MAGERR_PARAMS
############################################################

class Mask(object):
    """
    Contains maps of completeness depth in magnitudes for multiple observing bands, and associated products.
    """
    def __init__(self, config, roi):
        self.config = Config(config)
        self.roi = roi
        filenames = self.config.getFilenames()
        catalog_pixels = self.roi.getCatalogPixels()

        self.mask_1 = MaskBand(filenames['mask_1'][catalog_pixels],self.roi)
        self.mask_2 = MaskBand(filenames['mask_2'][catalog_pixels],self.roi)
        self._fracRoiSparse()

        self.minimum_solid_angle = self.config.params['mask']['minimum_solid_angle'] # deg^2

        # FIXME: Need to parallelize CMD and MMD formulation
        self._solidAngleCMD()
        self._pruneCMD(self.minimum_solid_angle)
        
        #self._solidAngleMMD()
        #self._pruneMMD(self.minimum_solid_angle)

        self._photometricErrors()

    @property
    def mask_roi_unique(self):
        """
        Assemble a set of unique magnitude tuples for the ROI
        """
        # There is no good inherent way in numpy to do this...
        # http://stackoverflow.com/q/16970982/

        # Also possible and simple:
        #return np.unique(zip(self.mask_1.mask_roi_sparse,self.mask_2.mask_roi_sparse))

        A = np.vstack([self.mask_1.mask_roi_sparse,self.mask_2.mask_roi_sparse]).T
        B = A[np.lexsort(A.T[::-1])]
        return B[np.concatenate(([True],np.any(B[1:]!=B[:-1],axis=1)))]

    @property
    def mask_roi_digi(self):
        """
        Get the index of the unique magnitude tuple for each pixel in the ROI.
        """
        # http://stackoverflow.com/q/24205045/#24206440
        A = np.vstack([self.mask_1.mask_roi_sparse,self.mask_2.mask_roi_sparse]).T
        B = self.mask_roi_unique

        AA = np.ascontiguousarray(A)
        BB = np.ascontiguousarray(B)
         
        dt = np.dtype((np.void, AA.dtype.itemsize * AA.shape[1]))
        a = AA.view(dt).ravel()
        b = BB.view(dt).ravel()
         
        idx = np.argsort(b)
        indices = np.searchsorted(b[idx],a)
        return idx[indices]

    @property
    def frac_annulus_sparse(self):
        return self.frac_roi_sparse[self.roi.pixel_annulus_cut]
     
    @property
    def frac_interior_sparse(self):
        return self.frac_roi_sparse[self.roi.pixel_interior_cut]

    def _fracRoiSparse(self):
        """
        Calculate an approximate pixel coverage fraction from the two masks.

        We have no way to know a priori how much the coverage of the
        two masks overlap in a give pixel. For example, masks that each
        have frac = 0.5 could have a combined frac = [0.0 to 0.5]. 
        The limits will be: 
          max:  min(frac1,frac2)
          min:  max((frac1+frac2)-1, 0.0)

        Sometimes we are lucky and our fracdet is actually already
        calculated for the two masks combined, so that the max
        condition is satisfied. That is what we will assume...
        """
        self.frac_roi_sparse = np.min([self.mask_1.frac_roi_sparse,self.mask_2.frac_roi_sparse],axis=0)
        return self.frac_roi_sparse

    def _solidAngleMMD(self):
        """
        Compute solid angle within the mask annulus (deg^2) as a
        function of mag_1 and mag_2
        """
        # Take upper corner of the magnitude bin
        mag_2,mag_1 = np.meshgrid(self.roi.bins_mag[1:],self.roi.bins_mag[1:])

        # Havent tested since adding fracdet
        unmasked_mag_1 = (self.mask_1.mask_annulus_sparse[:,np.newaxis]>mag_1[:,np.newaxis])
        unmasked_mag_2 = (self.mask_2.mask_annulus_sparse[:,np.newaxis]>mag_2[:,np.newaxis])
        n_unmasked_pixels = (unmasked_mag_1*unmasked_mag_2*self.frac_annulus_sparse).sum(axis=1)

        self.solid_angle_mmd = self.roi.area_pixel * n_unmasked_pixels

        if self.solid_angle_mmd.sum() == 0:
            msg = "Mask annulus contains no solid angle."
            logger.error(msg)
            raise Exception(msg)

    def _pruneMMD(self, minimum_solid_angle):
        """
        Remove regions of magnitude-magnitude space where the unmasked solid angle is
        statistically insufficient to estimate the background.

        INPUTS:
            solid_angle[1]: minimum solid angle (deg^2)
        """

        logger.info('Pruning mask based on minimum solid angle of %.2f deg^2'%(minimum_solid_angle))

        solid_angle_mmd = self.solid_angle_mmd*(self.solid_angle_mmd > minimum_solid_angle)
        if solid_angle_mmd.sum() == 0:
            msg = "Pruned mask contains no solid angle."
            logger.error(msg)
            raise ValueError(msg)

        self.solid_angle_mmd = solid_angle_mmd

        # Compute which magnitudes the clipping correspond to
        index_mag_1, index_mag_2 = np.nonzero(self.solid_angle_mmd)
        self.mag_1_clip = self.roi.bins_mag[1:][np.max(index_mag_1)]
        self.mag_2_clip = self.roi.bins_mag[1:][np.max(index_mag_2)]

        logger.info('Clipping mask 1 at %.2f mag'%(self.mag_1_clip) )
        logger.info('Clipping mask 2 at %.2f mag'%(self.mag_2_clip) )
        self.mask_1.mask_roi_sparse = np.clip(self.mask_1.mask_roi_sparse, 0., self.mag_1_clip)
        self.mask_2.mask_roi_sparse = np.clip(self.mask_2.mask_roi_sparse, 0., self.mag_2_clip)

    def _solidAngleCMD(self):
        """
        Compute solid angle within the mask annulus (deg^2) as a
        function of color and magnitude.
        """

        self.solid_angle_cmd = np.zeros([len(self.roi.centers_mag),
                                            len(self.roi.centers_color)])

        for index_mag in np.arange(len(self.roi.centers_mag)):
            for index_color in np.arange(len(self.roi.centers_color)):
                # mag and color at bin center
                mag = self.roi.centers_mag[index_mag]
                color = self.roi.centers_color[index_color]

                if self.config.params['catalog']['band_1_detection']:
                    # Evaluating at the center of the color-mag bin, be consistent!
                    #mag_1 = self.roi.centers_mag[index_mag]
                    #color = self.roi.centers_color[index_color]
                    #mag_2 = mag_1 - color
                    # Evaluating at corner of the color-mag bin, be consistent!
                    mag_1 = mag + (0.5 * self.roi.delta_mag)
                    mag_2 = mag - color + (0.5 * self.roi.delta_color)
                else:
                    # Evaluating at the center of the color-mag bin, be consistent!
                    #mag_2 = self.roi.centers_mag[index_mag]
                    #color = self.roi.centers_color[index_color]
                    #mag_1 = mag_2 + color
                    # Evaluating at corner of the color-mag bin, be consistent!
                    mag_1 = mag + color + (0.5 * self.roi.delta_color)
                    mag_2 = mag + (0.5 * self.roi.delta_mag)

                # ADW: Is there a problem here?
                #self.solid_angle_cmd[index_mag, index_color] = self.roi.area_pixel * np.sum((self.mask_1.mask > mag_1) * (self.mask_2.mask > mag_2))

                # ADW: I think we want to keep pixels that are >= mag
                unmasked_mag_1 = (self.mask_1.mask_annulus_sparse >= mag_1)
                unmasked_mag_2 = (self.mask_2.mask_annulus_sparse >= mag_2)
                n_unmasked_pixels = np.sum(unmasked_mag_1*unmasked_mag_2*self.frac_annulus_sparse)

                #n_unmasked_pixels = np.sum((self.mask_1.mask_annulus_sparse > mag_1) \
                #                               * (self.mask_2.mask_annulus_sparse > mag_2))

                self.solid_angle_cmd[index_mag, index_color] = self.roi.area_pixel * n_unmasked_pixels
        if self.solid_angle_cmd.sum() == 0:
            msg = "Mask annulus contains no solid angle."
            logger.error(msg)
            raise Exception(msg)

        return self.solid_angle_cmd


    def _solidAngleCMD(self):
        """
        Compute solid angle within the mask annulus (deg^2) as a
        function of color and magnitude.

        Returns:
        --------
        solid_angle_cmd : 2d array
        """

        self.solid_angle_cmd = np.zeros([len(self.roi.centers_mag),
                                            len(self.roi.centers_color)])

        idx_mag,idx_color=np.where(self.solid_angle_cmd == 0)
        mag = self.roi.centers_mag[idx_mag]
        color = self.roi.centers_color[idx_color]

        if self.config.params['catalog']['band_1_detection']:
            # Evaluating at corner of the color-mag bin, be consistent!
            mag_1 = mag + (0.5 * self.roi.delta_mag)
            mag_2 = mag - color + (0.5 * self.roi.delta_color)
        else:
            # Evaluating at corner of the color-mag bin, be consistent!
            mag_1 = mag + color + (0.5 * self.roi.delta_color)
            mag_2 = mag + (0.5 * self.roi.delta_mag)

        n_unmasked_pixels = np.zeros_like(mag)
        for i in np.arange(len(mag_1)):
            unmasked_mag_1 = (self.mask_1.mask_annulus_sparse >= mag_1[i])
            unmasked_mag_2 = (self.mask_2.mask_annulus_sparse >= mag_2[i])
            n_unmasked_pixels[i] = np.sum(unmasked_mag_1 * unmasked_mag_2 *
                                          self.frac_annulus_sparse)

        self.solid_angle_cmd[idx_mag, idx_color] = self.roi.area_pixel * n_unmasked_pixels
        if self.solid_angle_cmd.sum() == 0:
            msg = "Mask annulus contains no solid angle."
            logger.error(msg)
            raise Exception(msg)

        return self.solid_angle_cmd
        
    def _pruneCMD(self, minimum_solid_angle):
        """
        Remove regions of color-magnitude space where the unmasked solid angle is
        statistically insufficient to estimate the background.

        ADW: Why are we clipping at the bin center instead of edge?

        INPUTS:
            solid_angle[1]: minimum solid angle (deg^2)
        """

        logger.info('Pruning mask based on minimum solid angle of %.2f deg^2'%(minimum_solid_angle))
        self.solid_angle_cmd *= self.solid_angle_cmd > minimum_solid_angle

        if self.solid_angle_cmd.sum() == 0:
            msg = "Pruned mask contains no solid angle."
            logger.error(msg)
            raise Exception(msg)

        # Compute which magnitudes the clipping correspond to
        index_mag, index_color = np.nonzero(self.solid_angle_cmd)
        mag = self.roi.centers_mag[index_mag]
        color = self.roi.centers_color[index_color]
        if self.config.params['catalog']['band_1_detection']:
            mag_1 = mag
            mag_2 = mag_1 - color
            self.mag_1_clip = np.max(mag_1) + (0.5 * self.roi.delta_mag)
            self.mag_2_clip = np.max(mag_2) + (0.5 * self.roi.delta_color)
        else:
            mag_2 = mag
            mag_1 = color + mag_2
            self.mag_1_clip = np.max(mag_1) + (0.5 * self.roi.delta_color)
            self.mag_2_clip = np.max(mag_2) + (0.5 * self.roi.delta_mag)

        logger.info('Clipping mask 1 at %.2f mag'%(self.mag_1_clip) )
        logger.info('Clipping mask 2 at %.2f mag'%(self.mag_2_clip) )
        self.mask_1.mask_roi_sparse = np.clip(self.mask_1.mask_roi_sparse, 0., self.mag_1_clip)
        self.mask_2.mask_roi_sparse = np.clip(self.mask_2.mask_roi_sparse, 0., self.mag_2_clip)
        

    def completeness(self, delta, method='step'):
        """
        Return the completeness as a function of magnitude.

        ADW: Eventually want a completeness mask to set overall efficiency.
        """
        delta = np.asarray(delta)
        if method == 'step':
            func = lambda delta: (delta > 0).astype(float)
        elif method == 'erf':
            # Trust the SDSS EDR???
            # 95% completeness: 
            def func(delta):
                # Efficiency at bright end (assumed to be 100%)
                e = 1.0
                # EDR says full width is ~0.5 mag
                width = 0.2 
                # This should be the halfway point in the curve
                return (e/2.0)*(1/np.sqrt(2*width))*(np.sqrt(2*width)-scipy.special.erf(-delta))
        elif method == 'flemming':
            # Functional form taken from Fleming et al. AJ 109, 1044 (1995)
            # http://adsabs.harvard.edu/abs/1995AJ....109.1044F
            # f = 1/2 [1 - alpha(V - Vlim)/sqrt(1 + alpha^2 (V - Vlim)^2)]
            # CAREFUL: This definition is for Vlim = 50% completeness
            def func(delta):
                alpha = 2.0
                return 0.5 * (1 - (alpha * delta)/np.sqrt(1+alpha**2 * delta**2))
        else:
            raise Exception('...')
        return func(delta)

    def _photometricErrors(self, catalog=None, n_per_bin=100):
        """
        Realistic photometric errors estimated from catalog objects and mask.
        Extend below the magnitude threshold with a flat extrapolation.
        """

        if catalog is None:
            # Simple proxy for photometric errors
            release = self.config['data']['release']
            band_1 = self.config['catalog'].get('mag_1_band')
            if not band_1: band_1 = self.config['isochrone']['mag_1_field']
            band_2 = self.config['catalog'].get('mag_2_band')
            if not band_2: band_2 = self.config['isochrone']['mag_2_field']
            
            DELMIN = 0.0
            pars_1 = MAGERR_PARAMS[release][band_1]
            
            def photo_err_1(delta):
                p = pars_1
                return np.clip(np.exp(p[0]*delta+p[1])+p[2], 0, np.exp(p[0]*(DELMIN)+p[1])+p[2])

            pars_2 = MAGERR_PARAMS[release][band_2]
            def photo_err_2(delta):
                p = pars_2
                return np.clip(np.exp(p[0]*delta+p[1])+p[2], 0, np.exp(p[0]*(DELMIN)+p[1])+p[2])

        else:
            catalog.spatialBin(self.roi)

            if len(catalog.mag_1) < n_per_bin:
                logger.warning("Catalog contains fewer objects than requested to calculate errors.")
                #n_per_bin = int(len(catalog.mag_1) / 3)
                return self._photometricErrors(catalog=None)
             
            # Band 1
            mag_1_thresh = self.mask_1.mask_roi_sparse[catalog.pixel_roi_index] - catalog.mag_1
            sorting_indices = np.argsort(mag_1_thresh)
            mag_1_thresh_sort = mag_1_thresh[sorting_indices]
            mag_err_1_sort = catalog.mag_err_1[sorting_indices]
             
            # ADW: Can't this be done with np.median(axis=?)
            mag_1_thresh_medians = []
            mag_err_1_medians = []
            for i in range(0, int(len(mag_1_thresh) / float(n_per_bin))):
                mag_1_thresh_medians.append(np.median(mag_1_thresh_sort[n_per_bin * i: n_per_bin * (i + 1)]))
                mag_err_1_medians.append(np.median(mag_err_1_sort[n_per_bin * i: n_per_bin * (i + 1)]))
             
            if mag_1_thresh_medians[0] > 0.:
                mag_1_thresh_medians = np.insert(mag_1_thresh_medians, 0, -99.)
                mag_err_1_medians = np.insert(mag_err_1_medians, 0, mag_err_1_medians[0])
             
            photo_err_1 = scipy.interpolate.interp1d(mag_1_thresh_medians, mag_err_1_medians,
                                                     bounds_error=False, fill_value=mag_err_1_medians[-1])
             
            # Band 2
            mag_2_thresh = self.mask_2.mask_roi_sparse[catalog.pixel_roi_index] - catalog.mag_2
            sorting_indices = np.argsort(mag_2_thresh)
            mag_2_thresh_sort = mag_2_thresh[sorting_indices]
            mag_err_2_sort = catalog.mag_err_2[sorting_indices]
             
            mag_2_thresh_medians = []
            mag_err_2_medians = []
            for i in range(0, int(len(mag_2_thresh) / float(n_per_bin))):
                mag_2_thresh_medians.append(np.median(mag_2_thresh_sort[n_per_bin * i: n_per_bin * (i + 1)]))
                mag_err_2_medians.append(np.median(mag_err_2_sort[n_per_bin * i: n_per_bin * (i + 1)]))
             
            if mag_2_thresh_medians[0] > 0.:
                mag_2_thresh_medians = np.insert(mag_2_thresh_medians, 0, -99.)
                mag_err_2_medians = np.insert(mag_err_2_medians, 0, mag_err_2_medians[0])
             
            photo_err_2 = scipy.interpolate.interp1d(mag_2_thresh_medians, mag_err_2_medians,
                                                     bounds_error=False, fill_value=mag_err_2_medians[-1])
             
        self.photo_err_1=photo_err_1
        self.photo_err_2=photo_err_2

        return self.photo_err_1, self.photo_err_2

    def backgroundMMD(self, catalog, method='cloud-in-cells', weights=None):
        """
        Generate an empirical background model in magnitude-magnitude space.
        
        Parameters
        ----------
        catalog: catalog object
        method:  method for estimated MMD
        weights: weights assigned to each catalog object

        Returns
        -------
        background: estimate of background magnitude-magnitude distribution
        """

        # Select objects in annulus
        cut_annulus = self.roi.inAnnulus(catalog.lon,catalog.lat)
        mag_1 = catalog.mag_1[cut_annulus]
        mag_2 = catalog.mag_2[cut_annulus]

        # Units are (deg^2)
        solid_angle = ugali.utils.binning.take2D(self.solid_angle_mmd, mag_2, mag_1,
                                                 self.roi.bins_mag, self.roi.bins_mag)

        # Weight each object before binning
        # Divide by solid angle and bin size in magnitudes to get number density 
        # [objs / deg^2 / mag^2]
        if weights is None:
            number_density = (solid_angle*self.roi.delta_mag**2)**(-1)
        else:
            number_density = weights*(solid_angle*self.roi.delta_mag**2)**(-1)

        method = str(method).lower()
        if method == 'cloud-in-cells':
            # Apply cloud-in-cells algorithm
            mmd_background = ugali.utils.binning.cloudInCells(mag_2,mag_1,
                                                              [self.roi.bins_mag,self.roi.bins_mag],
                                                              weights=number_density)[0]
        elif method == 'bootstrap':
            # Not implemented; see catalog.bootstrap
            raise ValueError("Bootstrap method not implemented")

        elif method == 'histogram':
            # Apply raw histogram
            mmd_background = np.histogram2d(mag_1,mag_2,bins=[self.roi.bins_mag,self.roi.bins_mag],
                                            weights=number_density)[0]

        elif method == 'kde':
            # Gridded kernel density estimator
            logger.warning("### KDE not implemented properly")
            mmd_background = ugali.utils.binning.kernelDensity(mag_2,mag_1,
                                                               [self.roi.bins_mag,self.roi.bins_mag],
                                                               weights=number_density)[0]
        elif method == 'uniform':
            logger.warning("### WARNING: Uniform CMD")
            hist = np.histogram2d(mag_1,mag_2,bins=[self.roi.bins_mag,self.roi.bins_mag], 
                                  weights=number_density)[0]
            mmd_background = np.mean(hist)*np.ones(hist.shape)
            observable = (self.solid_angle_mmd > self.minimum_solid_angle)
            mmd_background *= observable
            return mmd_background
        else:
            raise ValueError("Unrecognized method: %s"%method)
        ## Account for the objects that spill out of the observable space
        ## But what about the objects that spill out to red colors??
        #for index_color in range(0, len(self.roi.centers_color)):
        #    for index_mag in range(0, len(self.roi.centers_mag)):
        #        if self.solid_angle_cmd[index_mag][index_color] < self.minimum_solid_angle:
        #            cmd_background[index_mag - 1][index_color] += cmd_background[index_mag][index_color]
        #            cmd_background[index_mag][index_color] = 0.
        #            break

        mmd_area = self.solid_angle_mmd*self.roi.delta_mag**2 # [deg^2 * mag^2]

        # ADW: This accounts for leakage to faint magnitudes
        # But what about the objects that spill out to red colors??
        # Maximum obsevable magnitude index for each color (uses the fact that
        # np.argmin returns first minimum (zero) instance found.
        # NOTE: More complicated maps may have holes causing problems

        observable = (self.solid_angle_mmd > self.minimum_solid_angle)
        index_mag_1 = observable.argmin(axis=0) - 1
        index_mag_2 = np.arange(len(self.roi.centers_mag))
        # Add the cumulative leakage back into the last bin of the CMD
        leakage = (mmd_background * ~observable).sum(axis=0)
        ### mmd_background[[index_mag_1,index_mag_2]] += leakage
        # Zero out all non-observable bins
        ### mmd_background *= observable

        # Avoid dividing by zero by setting empty bins to the value of the 
        # minimum filled bin of the CMD. This choice is arbitrary and 
        # could be replaced by a static minimum, some fraction of the 
        # CMD maximum, some median clipped minimum, etc. However, should 
        # be robust against outliers with very small values.
        min_mmd_background = max(mmd_background[mmd_background > 0.].min(),
                                 1e-4*mmd_background.max())
        mmd_background[observable] = mmd_background[observable].clip(min_mmd_background)

        ### # ADW: This is a fudge factor introduced to renormalize the CMD
        ### # to the number of input stars in the annulus. While leakage
        ### # will still smooth the distribution, it shouldn't result in 
        ### fudge_factor = len(mag) / float((cmd_background*cmd_area).sum())
        ### cmd_background *= fudge_factor

        return mmd_background

    def backgroundCMD(self, catalog, mode='cloud-in-cells', weights=None):
        """
        Generate an empirical background model in color-magnitude space.
        
        Parameters
        ----------
        catalog: catalog object
        method:  method for estimated MMD
        weights: weights assigned to each catalog object

        Returns
        -------
        background: estimate of background color-magnitude distribution
        """

        # Select objects in annulus
        cut_annulus = self.roi.inAnnulus(catalog.lon,catalog.lat)
        color = catalog.color[cut_annulus]
        mag   = catalog.mag[cut_annulus]

        # Units are (deg^2)
        solid_angle = ugali.utils.binning.take2D(self.solid_angle_cmd, color, mag,
                                                 self.roi.bins_color, self.roi.bins_mag)

        # Weight each object before binning
        # Divide by solid angle and bin size in magnitudes to get number density 
        # [objs / deg^2 / mag^2]
        if weights is None:
            number_density = (solid_angle*self.roi.delta_color*self.roi.delta_mag)**(-1)
        else:
            number_density = weights*(solid_angle*self.roi.delta_color*self.roi.delta_mag)**(-1)

        mode = str(mode).lower()
        if mode == 'cloud-in-cells':
            # Apply cloud-in-cells algorithm
            cmd_background = ugali.utils.binning.cloudInCells(color,mag,
                                                              [self.roi.bins_color,self.roi.bins_mag],
                                                              weights=number_density)[0]
        elif mode == 'bootstrap':
            # Not implemented; see catalog.bootstrap
            raise ValueError("Bootstrap mode not implemented")

        elif mode == 'histogram':
            # Apply raw histogram
            cmd_background = np.histogram2d(mag,color,bins=[self.roi.bins_mag,self.roi.bins_color],
                                            weights=number_density)[0]

        elif mode == 'kde':
            # Gridded kernel density estimator
            logger.warning("### KDE not implemented properly")
            cmd_background = ugali.utils.binning.kernelDensity(color,mag,
                                                               [self.roi.bins_color,self.roi.bins_mag],
                                                               weights=number_density)[0]
        elif mode == 'uniform':
            logger.warning("### WARNING: Uniform CMD")
            hist = np.histogram2d(mag,color,bins=[self.roi.bins_mag,self.roi.bins_color], weights=number_density)[0]
            cmd_background = np.mean(hist)*np.ones(hist.shape)
            observable = (self.solid_angle_cmd > self.minimum_solid_angle)
            cmd_background *= observable
            return cmd_background
        else:
            raise ValueError("Unrecognized mode: %s"%mode)
        ## Account for the objects that spill out of the observable space
        ## But what about the objects that spill out to red colors??
        #for index_color in range(0, len(self.roi.centers_color)):
        #    for index_mag in range(0, len(self.roi.centers_mag)):
        #        if self.solid_angle_cmd[index_mag][index_color] < self.minimum_solid_angle:
        #            cmd_background[index_mag - 1][index_color] += cmd_background[index_mag][index_color]
        #            cmd_background[index_mag][index_color] = 0.
        #            break

        cmd_area = self.solid_angle_cmd*self.roi.delta_color*self.roi.delta_mag # [deg^2 * mag^2]

        # ADW: This accounts for leakage to faint magnitudes
        # But what about the objects that spill out to red colors??
        # Maximum obsevable magnitude index for each color (uses the fact that
        # np.argmin returns first minimum (zero) instance found.
        # NOTE: More complicated maps may have holes causing problems

        observable = (self.solid_angle_cmd > self.minimum_solid_angle)
        index_mag = observable.argmin(axis=0) - 1
        index_color = np.arange(len(self.roi.centers_color))
        # Add the cumulative leakage back into the last bin of the CMD
        leakage = (cmd_background * ~observable).sum(axis=0)
        cmd_background[(index_mag,index_color)] += leakage
        # Zero out all non-observable bins
        cmd_background *= observable

        # Avoid dividing by zero by setting empty bins to the value of the 
        # minimum filled bin of the CMD. This choice is arbitrary and 
        # could be replaced by a static minimum, some fraction of the 
        # CMD maximum, some median clipped minimum, etc. However, should 
        # be robust against outliers with very small values.
        min_cmd_background = max(cmd_background[cmd_background > 0.].min(),
                                 1e-4*cmd_background.max())
        cmd_background[observable] = cmd_background[observable].clip(min_cmd_background)

        ### # ADW: This is a fudge factor introduced to renormalize the CMD
        ### # to the number of input stars in the annulus. While leakage
        ### # will still smooth the distribution, it shouldn't result in 
        ### fudge_factor = len(mag) / float((cmd_background*cmd_area).sum())
        ### cmd_background *= fudge_factor

        return cmd_background

    def restrictCatalogToObservableSpaceMMD(self, catalog):
        """
        Retain only the catalog objects which fall within the observable (i.e., unmasked) space.

        Parameters:
        catalog: a Catalog object
        Returns:
        sel    : boolean selection array where True means the object would be observable (i.e., unmasked).

        ADW: Careful, this function is fragile! The selection here should
             be the same as isochrone.observableFraction space. However,
             for technical reasons it is faster to do the calculation with
             broadcasting there.
        """

        # ADW: This creates a slope in color-magnitude space near the magnitude limit
        # i.e., if color=g-r then you can't have an object with g-r=1 and mag_r > mask_r-1
        # Depending on which is the detection band, this slope will appear at blue
        # or red colors. When it occurs at blue colors, it effects very few objects.
        # However, when occuring for red objects it can cut many objects. It is 
        # unclear that this is being correctly accounted for in the likelihood

        catalog.spatialBin(self.roi)
        sel_roi = (catalog.pixel_roi_index >= 0) # Objects outside ROI have pixel_roi_index of -1
        sel_mag_1 = catalog.mag_1 < self.mask_1.mask_roi_sparse[catalog.pixel_roi_index]
        sel_mag_2 = catalog.mag_2 < self.mask_2.mask_roi_sparse[catalog.pixel_roi_index]

        # and are located in the region of mag-mag space where background can be estimated
        sel_mmd = ugali.utils.binning.take2D(self.solid_angle_mmd,
                                             catalog.mag_2, catalog.mag_1,
                                             self.roi.bins_mag, self.roi.bins_mag) > 0.

        sel = np.all([sel_roi,sel_mag_1,sel_mag_2,sel_mmd], axis=0)
        return sel

    def restrictCatalogToObservableSpaceCMD(self, catalog):
        """
        Retain only the catalog objects which fall within the
        observable (i.e., unmasked) space.  NOTE: This returns a
        *selection* (i.e., objects are retained if the value of the
        output array is True).

        Parameters:
        catalog: a Catalog object
        Returns:
        sel    : boolean selection array where True means the object would be observable (i.e., unmasked).

        ADW: Careful, this function is fragile! The selection here should
             be the same as isochrone.observableFraction space. However,
             for technical reasons it is faster to do the calculation with
             broadcasting there.
        """

        # ADW: This creates a slope in color-magnitude space near the magnitude limit
        # i.e., if color=g-r then you can't have an object with g-r=1 and mag_r > mask_r-1
        # Depending on which is the detection band, this slope will appear at blue
        # or red colors. When it occurs at blue colors, it effects very few objects.
        # However, when occuring for red objects it can cut many objects. It is 
        # unclear that this is being correctly accounted for in the likelihood

        ### # Check that the objects fall in the color-magnitude space of the ROI
        ### # ADW: I think this is degenerate with the cut_cmd
        ### sel_mag = np.logical_and(catalog.mag > self.roi.bins_mag[0],
        ###                             catalog.mag < self.roi.bins_mag[-1])
        ### sel_color = np.logical_and(catalog.color > self.roi.bins_color[0],
        ###                               catalog.color < self.roi.bins_color[-1])

        # and are observable in the ROI-specific mask for both bands
        #if not hasattr(catalog, 'pixel_roi_index'): # TODO: An attempt to save computations, but not robust
        #    catalog.spatialBin(self.roi)
        catalog.spatialBin(self.roi)
        sel_roi = (catalog.pixel_roi_index >= 0) # Objects outside ROI have pixel_roi_index of -1
        sel_mag_1 = catalog.mag_1 < self.mask_1.mask_roi_sparse[catalog.pixel_roi_index]
        sel_mag_2 = catalog.mag_2 < self.mask_2.mask_roi_sparse[catalog.pixel_roi_index]

        # and are located in the region of color-magnitude space where background can be estimated
        sel_cmd = ugali.utils.binning.take2D(self.solid_angle_cmd,
                                             catalog.color, catalog.mag,
                                             self.roi.bins_color, self.roi.bins_mag) > 0.

        sel = np.all([sel_roi,sel_mag_1,sel_mag_2,sel_cmd], axis=0)
        return sel
    
    # FIXME: Need to parallelize CMD and MMD formulation
    restrictCatalogToObservableSpace = restrictCatalogToObservableSpaceCMD

############################################################

class MaskBand(object):
    """
    Map of liming magnitude for a single observing band.
    """

    def __init__(self, infiles, roi):
        """
        Parameters:
        -----------
        infiles : list of sparse healpix mask files
        roi : roi object

        Returns:
        --------
        mask : MaskBand object
        """
        self.roi = roi
        self.config = self.roi.config

        # ADW: It's overkill to make the full map just to slim it
        # down, but we don't have a great way to go from map pixels to
        # roi pixels.
        nside,pixel,maglim = healpix.read_partial_map(infiles,column='MAGLIM')
        self.nside = nside

        # Sparse maps of pixels in various ROI regions
        self.mask_roi_sparse = maglim[self.roi.pixels] 

        # Try to get the detection fraction
        self.frac_roi_sparse = (self.mask_roi_sparse > 0)
        try: 
            logger.debug("Reading FRACDET...")
            nside,pixel,frac=healpix.read_partial_map(infiles,column='FRACDET')
            # This clipping might gloss over bugs...
            fractype = self.config['mask'].get('fractype','binary')
            fracmin = self.config['mask'].get('fracmin',0.5)
            if fractype == 'binary':
                frac = np.where(frac < fracmin, 0.0, 1.0)
            elif fractype == 'full':
                frac = np.where(frac < fracmin, 0.0, frac)
            elif not fractype:
                pass
            else:
                msg = "Unrecognized fractype: %s"%fractype
                logger.warn(msg)
                
            self.frac_roi_sparse = np.clip(frac[self.roi.pixels],0.0,1.0)
        except ValueError as e:
            # No detection fraction present
            msg = "No 'FRACDET' column found in masks; assuming FRACDET = 1.0"
            logger.warn(msg)

        # Explicitly zero the maglim of pixels with fracdet < fracmin
        self.mask_roi_sparse[self.frac_roi_sparse == 0] = 0.0

    #ADW: Safer and more robust (though slightly slower)
    @property
    def mask_annulus_sparse(self):
        return self.mask_roi_sparse[self.roi.pixel_annulus_cut]
     
    @property
    def mask_interior_sparse(self):
        return self.mask_roi_sparse[self.roi.pixel_interior_cut]

    @property
    def mask_roi_unique(self):
        return np.unique(self.mask_roi_sparse)

    @property
    def mask_roi_digi(self):
        """ Mapping from unique pixels to roi pixels """
        return np.digitize(self.mask_roi_sparse,bins=self.mask_roi_unique)-1

    @property
    def frac_annulus_sparse(self):
        return self.frac_roi_sparse[self.roi.pixel_annulus_cut]
     
    @property
    def frac_interior_sparse(self):
        return self.frac_roi_sparse[self.roi.pixel_interior_cut]

    def completeness(self, mags, method='step'):
        """
        Return the completeness as a function of magnitude.

        ADW: Eventually want a completeness mask to set overall efficiency.
        """
        if method == 'step':
            func = lambda x: (x < self.mask_roi_unique[:,np.newaxis]).astype(float)
        elif method == 'erf':
            # Trust the ERD???
            # 95% completeness: 
            def func(x):
                # Efficiency at bright end (assumed to be 100%)
                e = 1.0
                # SDSS EDR says full width is ~0.5 mag
                width = 0.2 
                # This should be the halfway point in the curve
                maglim = self.mask_roi_unique[:,np.newaxis]
                return (e/2.0)*(1/np.sqrt(2*width))*(np.sqrt(2*width)-scipy.special.erf((x-maglim)))
        else:
            raise Exception('...')

        return func(mags)
        
    def depth(self, lon, lat):
        """
        Magnitude limit at given image coordinates.
        """
        pass

class CoverageBand(object):
    """
    Map of coverage fraction for a single observing band.
    """

    def __init__(self, infiles, roi):
        """
        Infile is a sparse HEALPix map fits file.
        """
        self.roi = roi
        mask = ugali.utils.skymap.readSparseHealpixMaps(infiles, field='COVERAGE')
        self.nside = hp.npix2nside(len(mask))
        # Sparse maps of pixels in various ROI regions
        self.mask_roi_sparse = mask[self.roi.pixels] 


############################################################

class SimpleMask(Mask):
    """
    Contains maps of completeness depth in magnitudes for multiple observing bands, and associated products.
    """
    def __init__(self, config, roi, maglim_1=23, maglim_2=23):
        self.config = Config(config)
        self.roi = roi

        self.mask_1 = SimpleMaskBand(maglim_1,self.roi)
        self.mask_2 = SimpleMaskBand(maglim_2,self.roi)
        
        self.minimum_solid_angle = self.config.params['mask']['minimum_solid_angle'] # deg^2

        # FIXME: Need to parallelize CMD and MMD formulation
        self._solidAngleCMD()
        self._pruneCMD(self.minimum_solid_angle)
        
        #self._solidAngleMMD()
        #self._pruneMMD(self.minimum_solid_angle)

        self._photometricErrors()

############################################################

class SimpleMaskBand(MaskBand):
    """
    Map of completeness depth in magnitudes for a single observing band.
    """

    def __init__(self, maglim, roi):
        """
        Infile is a sparse HEALPix map fits file.
        """
        self.roi = roi
        mask = maglim*np.ones(hp.nside2npix(self.roi.config['coords']['nside_pixel']))
        self.nside = hp.npix2nside(len(mask))
        # Sparse maps of pixels in various ROI regions
        self.mask_roi_sparse = mask[self.roi.pixels] 

############################################################



"""
This is the Ultra-faint Galaxy Likelihood (UGaLi) software package.
"""
__author__ = "Keith Bechtol & Alex Drlica-Wagner"
__email__ = "bechtol@kicp.uchicago.edu, kadrlica@fnal.gov"

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

# ADW: Is this a good idea?
#import ugali.analysis.isochrone as isochrone
#from ugali import isochrone

#import ugali.analysis.kernel as kernel
#import ugali.analysis.source as source

# Hack for backward compatibitility with: ugali.analysis.isochrone
#sys.modules['ugali.analysis.isochrone'] = __import__('ugali.isochrone')



#!/usr/bin/env python
"""
Run MCMC follow-up on target parameters.

The default parellel execution by emcee uses multiprocess and pickeling
the function that is being evaluated. Things become a bit complicated 
because you can't pickle instance methods.
Steven Bethard has a solution registering a method with pickle:
http://bytes.com/topic/python/answers/552476-why-cant-you-pickle-instancemethods#edit2155350

Another simple solution is to create a plain method linked to a static method.
http://stackoverflow.com/questions/21111106/cant-pickle-static-method-multiprocessing-python

We choose the second option here.
"""

import os,sys
from collections import OrderedDict as odict

import numpy
import numpy as np
import numpy.lib.recfunctions as recfuncs
import scipy.stats

import yaml

import ugali.utils.stats
from ugali.utils.stats import Samples
import ugali.analysis.scan
from ugali.utils.config import Config
from ugali.analysis.kernel import kernelFactory
from ugali.analysis.loglike import createSource, createLoglike
from ugali.analysis.prior import UniformPrior, InversePrior

from ugali.utils.logger import logger
from ugali.utils.projector import dist2mod,mod2dist,gal2cel,gal2cel_angle

try:
    from mpi4py import rc
    rc.initialize = False
except ImportError:
    pass

#ADW: Why did I need to get rid of this?
#import multiprocessing

import emcee

global mcmc,niter

niter = 0
mcmc = None

def lnprob(theta):
    return mcmc.lnprob(theta)

class MCMC(object):
    """
    This object creates the loglike object from a source model along
    with a log prior and passes them to emcee with some flow control
    parameters.
    """

    def __init__(self, config, loglike):
        self.config = Config(config)
        self.nsamples = self.config['mcmc'].get('nsamples',100)
        self.nthreads = self.config['mcmc'].get('nthreads',16)
        #self.nthreads = multiprocessing.cpu_count()
        self.nchunk = self.config['mcmc'].get('nchunk',25)
        self.nwalkers = self.config['mcmc'].get('nwalkers',50)
        self.nburn = self.config['mcmc'].get('nburn',10)

        self.loglike = loglike
        self.source = self.loglike.source
        self.params = list(self.source.get_free_params().keys())
        self.samples = None

        self.priors = odict(list(zip(self.params,
                                     len(self.params)*[UniformPrior()])))
        self.priors['extension'] = InversePrior()

        self.pool = None

    def __str__(self):
        ret = "%s:\n"%self.__class__.__name__
        ret += str(self.loglike)
        return ret

    def __call__(self, theta):
        self.lnprob(theta)

    def get_mle(self):
        """ Get the values of the source parameters. """
        #return self.grid.mle()
        mle = self.source.get_params()
        # FIXME: For composite isochrones
        if 'age' not in mle:
            mle['age'] = np.average(self.source.isochrone.age)
        if 'metallicity' not in mle:
            mle['metallicity'] = np.average(self.source.isochrone.metallicity)
            
        return mle
 
    def get_std(self):
        """
        Somewhat arbitrarily assign standard deviation to seed points.
        """
        mle = self.get_mle()
 
        std = odict([
            ('richness',0.1*mle['richness']), # delta_r (10% of max)
            ('lon',0.01),                     # delta_lon (deg)
            ('lat',0.01),                     # delta_lat (deg)
            ('distance_modulus',0.1),         # delta_mu                
            ('extension',0.01),               # delta_ext (deg)
            ('ellipticity',0.1),              # delta_e 
            ('position_angle',15.0),          # delta_pa (deg)
            ('age',0.5),                      # delta_age (Gyr)
            ('metallicity',0.0001),           # delta_z
        ])
        return std
 
    def get_ball(self, params, size=1):
        mle = self.get_mle() 
        std = self.get_std()
                                             
        p0 = np.array([mle[k] for k in params])
        s0 = np.array([std[k] for k in params])
        ball = emcee.utils.sample_ball(p0,s0,size) 

        # Set points outside the bounds to the mle estimate
        for i,param in enumerate(params):
            bounds = self.source.params[param].bounds
            outside = (ball[:,i] < bounds[0]) | (ball[:,i] > bounds[1])
            ball[:,i][outside] = p0[i]

        return ball

    def get_uniform(self, params, size=1):
        # Placeholder for starting walkers over a uniform distribution
        raise Exception("'get_uniform' is not implemented")

        ball = np.ones((len(params),size),dtype=float)
        for i,param in enumerate(params):
            bounds = self.source.params[param].bounds
            # How do we want to deal with unbounded (inf) parameters?
            ball[:,i] = np.random.uniform(bounds[0],bounds[1],size)
        return ball

    def lnlike(self, theta):
        """ Logarithm of the likelihood """
        params,loglike = self.params,self.loglike
        kwargs = dict(list(zip(params,theta)))
        try:
            lnlike = loglike.value(**kwargs)
        except ValueError as AssertionError:
            lnlike = -np.inf
        return lnlike
 
    def lnprior(self,theta):
        """ Logarithm of the prior """
        params,priors = self.params,self.priors
        kwargs = dict(list(zip(params,theta)))
        err = np.seterr(invalid='raise')
        try:
            lnprior = np.sum(np.log([priors[k](v) for k,v in list(kwargs.items())]))
        except (FloatingPointError,ValueError):
            lnprior = -np.inf
        np.seterr(**err)
        return lnprior

    def lnprob(self,theta):
        """ Logarithm of the probability """
        global niter
        params,priors,loglike = self.params,self.priors,self.loglike
        # Avoid extra likelihood calls with bad priors
        _lnprior = self.lnprior(theta)
        if np.isfinite(_lnprior):
            _lnlike = self.lnlike(theta)
        else:
            _lnprior = -np.inf
            _lnlike = -np.inf

        _lnprob = _lnprior + _lnlike
     
        if (niter%100==0):
            msg = "%i function calls ...\n"%niter
            msg+= ', '.join('%s: %.3f'%(k,v) for k,v in zip(params,theta))
            msg+= '\nlog(like): %.3f, log(prior): %.3f'%(_lnprior,_lnlike)
            logger.debug(msg)
        niter+=1
        return _lnprob

    def run(self, params=None, outfile=None):
        # Initailize the likelihood to maximal value
        mle =self.get_mle()
        msg = "Setting inital values..."
        for k,v in mle.items():
            msg+='\n  %s : %s'%(k,v)
        logger.info(msg)
 
        self.loglike.set_params(**mle)
        self.loglike.sync_params()

        logger.info(str(self.loglike))
 
        if params is not None: self.params = params
        if len(self.params) == 0:
            raise Exception("No free parameters to fit.")
            
        params   = self.params
        nwalkers = self.nwalkers
        nsamples = self.nsamples
        nthreads = self.nthreads
        nburn    = self.nburn
        ndim = len(self.params)
        
        logger.info("Running MCMC chain...")

        p0 = self.get_ball(self.params,nwalkers)

        kwargs = dict(threads=nthreads,pool=self.pool)
        self.sampler = emcee.EnsembleSampler(nwalkers,ndim,lnprob,**kwargs)

        # Burn the requested number of entries
        logger.info("Burning %i steps..."%nburn)
        pos,prob,state = self.sampler.run_mcmc(p0,nburn)
        self.sampler.reset() 

        # Chain is shape (nwalkers,nsteps,nparams)
        # Samples is shape (nwalkers*nsteps,nparams):
        #for i,result in enumerate(self.sampler.sample(p0,iterations=nsamples)):
        for i,result in enumerate(self.sampler.sample(pos,prob,state,iterations=nsamples)):
            steps = i+1
            if steps%10 == 0: logger.info("%i steps ..."%steps)
            self.chain = self.sampler.chain
            if (i==0) or (steps%self.nchunk==0):
                samples = self.chain.reshape(-1,len(self.params),order='F')
                self.samples = Samples(samples.T,names=self.params)
                if outfile is not None: 
                    logger.info("Writing %i steps to %s..."%(steps,outfile))
                    self.write_samples(outfile)
 
        samples = self.chain.reshape(-1,len(self.params),order='F')
        self.samples = Samples(samples.T,names=self.params)

        if outfile is not None: self.write_samples(outfile)
 
    def write_samples(self,filename):
        np.save(filename,self.samples)
 
    def load_samples(self,filename):
        self.samples = Samples(filename)
 
    def load_srcmdl(self,filename,section='source'):
        self.source.load(filename,section)

    def write_srcmdl(self,filename,section='source'):
        source = dict()
        source[section] = self.source.todict()
        
        out = open(filename,'w')
        out.write(yaml.dump(source))
        out.close()

    def __getstate__(self):
        # Remove the pool from the pickeled object
        #http://stackoverflow.com/a/25385582/4075339
        self_dict = self.__dict__.copy()

        if 'pool' in self_dict:    del self_dict['pool']
        if 'sampler' in self_dict: del self_dict['sampler']
            
        return self_dict

    def __setstate__(self, state):
        self.__dict__.update(state)

# Move to factory?
def createMCMC(config,srcfile,section='source',samples=None):
    """ Create an MCMC instance """

    source = ugali.analysis.source.Source()
    source.load(srcfile,section=section)
    loglike = ugali.analysis.loglike.createLoglike(config,source)

    mcmc = MCMC(config,loglike)
    if samples is not None:
        mcmc.load_samples(samples)

    return mcmc

if __name__ == "__main__":
    import ugali.utils.parser
    description = "Script for running MCMC followup"
    parser = ugali.utils.parser.Parser(description=description)
    parser.add_config()
    parser.add_debug()
    parser.add_verbose()
    #parser.add_coords(required=True)
    parser.add_coords()
    parser.add_name()
    parser.add_argument('--srcmdl',help='Source model file')
    parser.add_argument('--grid',action='store_true',help='Grid search for intial parameters')
    parser.add_argument('outfile',default=None,help="Output file name")
                        
    opts = parser.parse_args()

    if opts.coords is not None and len(opts.coords) != 1: 
        raise Exception('Must specify exactly one coordinate.')
    
    #try:
    #    # Initialize the MPI-based pool used for parallelization.
    #    from emcee.utils import MPIPool
    #    pool = MPIPool(loadbalance=True)
    #except (ImportError,ValueError) as e:
    #    logger.warn(e.message)
    #    pool = None
    # 
    #if pool and not pool.is_master():
    #    # Wait for instructions from the master process.
    #    #logger.setLevel(logger.CRITICAL)
    #    pool.wait()
    #    sys.exit(0)
    #pool = None

    config = Config(opts.config)

    outfile = opts.outfile
    samfile = outfile                          # samples
    srcfile = outfile.replace('.npy','.yaml')  # srcmdl and results
    memfile = outfile.replace('.npy','.fits')  # membership
    resfile = srcfile                          # results file
    
    source = ugali.analysis.loglike.createSource(config,section='mcmc')
    source.name = opts.name

    if opts.srcmdl is not None:
        source.load(opts.srcmdl,section=opts.name)
    if opts.coords:
        lon,lat,radius = opts.coords[0]
        if config['coords']['coordsys'].lower() == 'gal':
            source.set_params(lon=lon,lat=lat)
        else:
            lon,lat = gal2cel(lon,lat)
            source.set_params(lon=lon,lat=lat)

    if config['mcmc'].get('params'):
        params = config['mcmc'].get('params')
        source.set_free_params(params)

    like = ugali.analysis.loglike.createLoglike(config,source)

    if opts.grid:
        grid = ugali.analysis.scan.GridSearch(config,like)
        grid.search()
        source.set_params(**grid.mle())

    params = list(source.get_free_params().keys())
    logger.info(source)
    mcmc = MCMC(config,like)

    logger.info("Writing %s..."%srcfile)
    mcmc.write_srcmdl(srcfile)

    """
    # DEBUGGING
    loglike = mcmc.loglike
    roi = loglike.roi
    mask = loglike.mask
    
    ra = np.linspace(354.3,354.4)
    dec = np.linspace(-63.274,-63.256)

    rr,dd = np.meshgrid(ra,dec)
    zz = [mcmc.lnprob([2500,r,d,0.02]) for r,d in zip(rr.flat,dd.flat)]
    zz = np.array(zz).reshape(rr.shape)
    import pylab as plt
    plt.ion()
    plt.pcolormesh(rr,dd,zz)
    import skymap
    smap =skymap.Skymap(projection='cass',lon_0=354.36,lat_0=-63.265,
                        llcrnrlon=354.3,urcrnrlon=354.4,
                        llcrnrlat=-63.2725,urcrnrlat=-63.2575,
                        celestial=False)
    smap.draw_hpxmap(mask.mask_1.frac_roi_sparse,
    import pdb; pdb.set_trace()
    """

    mcmc.run(params,samfile)

    logger.info("Writing %s..."%srcfile)
    from ugali.analysis.results import write_results
    write_results(resfile,config,srcfile,samfile)

    logger.info("Writing %s..."%memfile)
    from ugali.analysis.loglike import write_membership
    write_membership(memfile,config,srcfile,section='source')



from __future__ import print_function

import sys
import os
import io

try: 
    from setuptools import setup, find_packages
    from setuptools.command.install import install as _install
except ImportError: 
    from distutils.core import setup
    from distutils.command.install import install as _install
    def find_packages():
        return ['ugali','ugali.analysis','ugali.config','ugali.observation',
                'ugali.preprocess','ugali.simulation','ugali.candidate',
                'ugali.utils']

import distutils.cmd

import versioneer
VERSION = versioneer.get_version()

NAME = 'ugali'
HERE = os.path.abspath(os.path.dirname(__file__))
URL = 'https://github.com/DarkEnergySurvey/ugali'
DESC = "Ultra-faint galaxy likelihood toolkit."
LONG_DESC = "%s\n%s"%(DESC,URL)
CLASSIFIERS = """\
Development Status :: 4 - Beta
Intended Audience :: Science/Research
Intended Audience :: Developers
License :: OSI Approved :: MIT License
Natural Language :: English
Operating System :: MacOS :: MacOS X
Operating System :: POSIX :: Linux
Programming Language :: Python :: 2
Programming Language :: Python :: 3
Topic :: Scientific/Engineering
Topic :: Scientific/Engineering :: Astronomy
Topic :: Scientific/Engineering :: Physics
"""

RELEASE_URL = URL+'/releases/download/v1.8.0'
UGALIDIR = os.getenv("UGALIDIR","$HOME/.ugali")
ISOSIZE = "~1MB" 
CATSIZE = "~20MB"
TSTSIZE = "~1MB"
# Could find file size dynamically, but it's a bit slow...
# int(urllib.urlopen(ISOCHRONES).info().getheaders("Content-Length")[0])/1024**2
SURVEYS = ['des','ps1','sdss','lsst']
MODELS = ['bressan2012','marigo2017','dotter2008','dotter2016']

class ProgressFileIO(io.FileIO):
    def __init__(self, path, *args, **kwargs):
        self._total_size = os.path.getsize(path)
        io.FileIO.__init__(self, path, *args, **kwargs)

    def read(self, size):
        count = self.tell()/size
        self.progress_bar(count,size,self._total_size)
        return io.FileIO.read(self, size)

    @staticmethod
    def progress_bar(count, block_size, total_size):
        block = 100*block_size/float(total_size)
        progress = count*block
        if progress % 5 < 1.01*block:
            msg = '\r[{:51}] ({:d}%)'.format(int(progress//2)*'='+'>',int(progress))
            sys.stdout.write(msg)
            sys.stdout.flush()

class TarballCommand(distutils.cmd.Command,object):
    """ Command for downloading data files """
    description = "install data files"
    user_options = [
        ('ugali-dir=',None,
         'path to install data files [default: %s]'%UGALIDIR),
        ('force','f',
         'force installation (overwrite any existing files)')
        ]
    boolean_options = ['force']
    release_url = RELEASE_URL
    _tarball = None
    _dirname = None

    def initialize_options(self):
        self.ugali_dir = os.path.expandvars(UGALIDIR)
        self.force = False
        # Not really the best way, but ok...
        self.tarball = self._tarball
        self.dirname = self._dirname
        
    def finalize_options(self):
        # Required by abstract base class
        pass

    @property
    def path(self):
        return os.path.join(self.ugali_dir,self.dirname)

    def check_exists(self):
        return os.path.exists(self.path)
        
    def install_tarball(self, tarball):
        try:
            from urllib.request import urlopen, urlretrieve
            from urllib.error import HTTPError
        except ImportError:
            from urllib import urlopen, urlretrieve
            from urllib2 import HTTPError

        import tarfile

        if not os.path.exists(self.ugali_dir):
            print("creating %s"%self.ugali_dir)
            os.makedirs(self.ugali_dir)
        os.chdir(self.ugali_dir)

        url = os.path.join(self.release_url,tarball)

        print("downloading %s..."%url)
        if urlopen(url).getcode() >= 400:
            raise Exception('url does not exist')

        urlretrieve(url,tarball,reporthook=ProgressFileIO.progress_bar)
        print('')
        if not os.path.exists(tarball):
            raise HTTPError()
            
        print("extracting %s..."%tarball)
        with tarfile.open(fileobj=ProgressFileIO(tarball),mode='r:gz') as tar:
            ## Check if the directory exists?
            #if os.path.exists(tar.next().name) and not self.force:
            #    print("directory found; skipping installation")
            tar.extractall()
            tar.close()
            print('')

        print("removing %s"%tarball)
        os.remove(tarball)

    def run(self):
        if self.dry_run:
            print("skipping data install")
            return
        
        if self.check_exists():
            print("found %s"%self.path)
            if self.force:
                print("overwriting directory")
            else:
                print("use '--force' to overwrite")
                return
       
        self.install_tarball(self.tarball)

class CatalogCommand(TarballCommand):
    """ Command for downloading catalog files """
    description = "install catalog files"
    _tarball = 'ugali-catalogs.tar.gz'
    _dirname = 'catalogs'

class TestsCommand(TarballCommand):
    """ Command for downloading catalog files """
    description = "install test data"
    _tarball = 'ugali-test-data.tar.gz'
    _dirname = 'testdata'

class IsochroneCommand(TarballCommand):
    """ Command for downloading isochrone files """
    description = "install isochrone files"
    user_options = TarballCommand.user_options + [
        ('survey=',None,
         'survey set [default: None]'),
        ('model=',None,
         'isochrone model [default: None]')
        ]
    _tarball = 'ugali-isochrones-tiny.tar.gz'
    _dirname = 'isochrones'

    def initialize_options(self):
        super(IsochroneCommand,self).initialize_options()
        self.survey = None
        self.model = None

    def finalize_options(self):
        super(IsochroneCommand,self).finalize_options()
        self._build_surveys()
        self._build_models()

    def _build_surveys(self):
        if self.survey is None:
            self.surveys = SURVEYS
        else:
            self.survey = self.survey.lower()
            if self.survey not in SURVEYS:
                raise Exception("unrecognized survey: '%s'"%self.survey)
            self.surveys = [self.survey]

    def _build_models(self):
        if self.model is None: 
            self.models = MODELS
        else:
            self.model = self.model.lower()
            if self.model not in MODELS:
                raise Exception("unrecognized model: '%s'"%self.model)
            self.models = [self.model]

    def run(self):
        if self.dry_run:
            print("skipping data install")
            return

        if (self.survey is None) and (self.model is None):
            self.tarball = self._tarball
            self.dirname = self._dirname
            super(IsochroneCommand,self).run()
            return
        
        for survey in self.surveys:
            for model in self.models:
                self.tarball = "ugali-%s-%s.tar.gz"%(survey,model)
                self.dirname = "isochrones/%s/%s"%(survey,model)
                super(IsochroneCommand,self).run()


class install(_install):
    """ 
    Subclass the setuptools 'install' class.
    """
    user_options = _install.user_options + [
        ('isochrones',None,"install isochrone files (%s)"%ISOSIZE),
        ('catalogs',None,"install catalog files (%s)"%CATSIZE),
        ('tests',None,"install test data (%s)"%TSTSIZE),
        ('ugali-dir=',None,"install file directory [default: %s]"%UGALIDIR),
    ]
    boolean_options = _install.boolean_options + ['isochrones','catalogs']

    def initialize_options(self):
        _install.initialize_options(self)
        self.ugali_dir = os.path.expandvars(UGALIDIR)
        self.isochrones = False
        self.catalogs = False
        self.tests = False

    def run(self):
        # run superclass install
        _install.run(self)

        # Could ask user whether they want to install isochrones, but 
        # pip filters sys.stdout, so the prompt never gets sent:
        # https://github.com/pypa/pip/issues/2732#issuecomment-97119093

        if self.isochrones: 
            self.install_isochrones()

        if self.catalogs: 
            self.install_catalogs()

        if self.tests:
            self.install_tests()

    def install_isochrones(self):
        """
        Call to isochrone install command:
        http://stackoverflow.com/a/24353921/4075339
        """
        cmd_obj = self.distribution.get_command_obj('isochrones')
        cmd_obj.force = self.force
        if self.ugali_dir: cmd_obj.ugali_dir = self.ugali_dir
        self.run_command('isochrones')

    def install_catalogs(self):
        """
        Call to catalog install command:
        http://stackoverflow.com/a/24353921/4075339
        """
        cmd_obj = self.distribution.get_command_obj('catalogs')
        cmd_obj.force = self.force
        if self.ugali_dir: cmd_obj.ugali_dir = self.ugali_dir
        self.run_command('catalogs')

    def install_tests(self):
        """
        Call to catalog install command:
        http://stackoverflow.com/a/24353921/4075339
        """
        cmd_obj = self.distribution.get_command_obj('tests')
        cmd_obj.force = self.force
        if self.ugali_dir: cmd_obj.ugali_dir = self.ugali_dir
        self.run_command('tests')

CMDCLASS = versioneer.get_cmdclass()
CMDCLASS['isochrones'] = IsochroneCommand
CMDCLASS['catalogs'] = CatalogCommand
CMDCLASS['tests'] = TestsCommand
CMDCLASS['install'] = install

setup(
    name=NAME,
    version=VERSION,
    cmdclass=CMDCLASS,
    url=URL,
    author='Keith Bechtol & Alex Drlica-Wagner',
    author_email='bechtol@wisc.edu, kadrlica@fnal.gov',
    scripts = [],
    install_requires=[
        'astropy',
        'matplotlib',
        'numpy >= 1.9.0',
        'scipy >= 0.14.0',
        'healpy >= 1.6.0',
        'fitsio >= 0.9.10',
        'emcee >= 2.1.0',
        'corner >= 1.0.0',
        'pyyaml >= 3.10',
    ],
    packages=find_packages(),
    description=DESC,
    long_description=LONG_DESC,
    platforms='any',
    classifiers = [_f for _f in CLASSIFIERS.split('\n') if _f]
)



#!/usr/bin/env python
"""
For creating healpix maps from catalogs.
"""
import os
from os.path import join
import shutil
import tempfile
import subprocess
from collections import Counter
from collections import OrderedDict as odict

import fitsio
import numpy as np
import numpy.lib.recfunctions as recfuncs
import healpy as hp
from scipy.interpolate import interp1d
from scipy.optimize import brentq

import ugali.utils.skymap
import ugali.utils.binning
from ugali.utils.projector import cel2gal, gal2cel
from ugali.utils.healpix import ang2pix, pix2ang, superpixel, read_map
from ugali.utils.shell import mkdir
from ugali.utils.logger import logger
from ugali.utils.config import Config
from ugali.utils.constants import MAGLIMS


class Maglims(object):
    """ Object for deriving magnitude limits from the catalog """

    def __init__(self, config):
        self.config = Config(config)
        self._setup()

    def _setup(self):
        self.nside_catalog = self.config['coords']['nside_catalog']
        self.nside_mask = self.config['coords']['nside_mask']
        self.nside_pixel = self.config['coords']['nside_pixel']
        
        self.filenames = self.config.getFilenames()
        
        self.footfile = self.config['data']['footprint']
        try: 
            self.footprint = fitsio.read(self.footfile)['I'].ravel()
        except:
            logger.warn("Couldn't open %s; will pass through."%self.footfile)
            self.footprint = self.footfile


    def run(self,field=None,simple=False,force=False):
        """
        Loop through pixels containing catalog objects and calculate
        the magnitude limit. This gets a bit convoluted due to all
        the different pixel resolutions...
        """
        if field is None: fields = [1,2]
        else:             fields = [field]
        for filenames in self.filenames.compress(~self.filenames.mask['catalog']).data:
            infile = filenames['catalog']
            for f in fields:
                outfile = filenames['mask_%i'%f]
                if os.path.exists(outfile) and not force:
                    logger.info("Found %s; skipping..."%outfile)
                    continue
                
                pixels,maglims=self.calculate(infile,f,simple)
                logger.info("Creating %s"%outfile)
                outdir = mkdir(os.path.dirname(outfile))
                data = odict()
                data['PIXEL']=pixels
                data['MAGLIM']=maglims.astype('f4')
                ugali.utils.healpix.write_partial_map(outfile,data,
                                                      self.nside_pixel)
                                                      

    def calculate(self, infile, field=1, simple=False):
        logger.info("Calculating magnitude limit from %s"%infile)

        #manglefile = self.config['mangle']['infile_%i'%field]
        #footfile = self.config['data']['footprint']
        #try: 
        #    footprint = fitsio.read(footfile)['I'].ravel()
        #except:
        #    logger.warn("Couldn't open %s; will try again."%footfile)
        #    footprint = footfile

        mag_column = self.config['catalog']['mag_%i_field'%field]
        magerr_column = self.config['catalog']['mag_err_%i_field'%field]

        # For simple maglims
        release = self.config['data']['release'].lower()
        band    = self.config['catalog']['mag_%i_band'%field]
        pixel_pix_name = 'PIX%i'%self.nside_pixel         

        # If the data already has a healpix pixel assignment then use it
        # Otherwise recalculate...
        try:
            data = fitsio.read(infile,columns=[pixel_pix_name])
        except ValueError as e:
            logger.info(str(e))
            columns=[self.config['catalog']['lon_field'],
                     self.config['catalog']['lat_field']]
            data = fitsio.read(infile,columns=columns)[columns]
            pix = ang2pix(self.nside_pixel,data[columns[0]],data[columns[1]])
            data = recfuncs.rec_append_fields(data,pixel_pix_name,pix)
            
        #mask_pixels = np.arange( hp.nside2npix(self.nside_mask), dtype='int')
        mask_maglims = np.zeros(hp.nside2npix(self.nside_mask))
         
        out_pixels = np.zeros(0,dtype='int')
        out_maglims = np.zeros(0)
         
        # Find the objects in each pixel
        pixel_pix = data[pixel_pix_name]
        mask_pix = ugali.utils.skymap.superpixel(pixel_pix,self.nside_pixel,self.nside_mask)
        count = Counter(mask_pix)
        pixels = sorted(count.keys())
        pix_digi = np.digitize(mask_pix,pixels).argsort()
        idx = 0
        min_num = 500
        signal_to_noise = 10.
        magerr_lim = 1/signal_to_noise
        for pix in pixels:
            # Calculate the magnitude limit in each pixel
            num = count[pix]
            objs = data[pix_digi[idx:idx+num]]
            idx += num
            if simple:
                # Set constant magnitude limits
                logger.debug("Simple magnitude limit for %s"%infile)
                mask_maglims[pix] = MAGLIMS[release][band]
            elif num < min_num:
                logger.info('Found <%i objects in pixel %i'%(min_num,pix))
                mask_maglims[pix] = 0
            else:
                mag = objs[mag_column]
                magerr = objs[magerr_column]
                # Estimate the magnitude limit as suggested by:
                # https://deswiki.cosmology.illinois.edu/confluence/display/DO/SVA1+Release+Document
                # (https://desweb.cosmology.illinois.edu/confluence/display/Operations/SVA1+Doc)
                maglim = np.median(mag[(magerr>0.9*magerr_lim)&(magerr<1.1*magerr_lim)])
         
                # Alternative method to estimate the magnitude limit by fitting median
                #mag_min, mag_max = mag.min(),mag.max()
                #mag_bins = np.arange(mag_min,mag_max,0.1) #0.1086?
                #x,y = ugali.utils.binning.binnedMedian(mag,magerr,mag_bins)
                #x,y = x[~np.isnan(y)],y[~np.isnan(y)]
                #magerr_med = interp1d(x,y)
                #mag0 = np.median(x) 
                #maglim = brentq(lambda a: magerr_med(a)-magerr_lim,x.min(),x.max(),disp=False)
                # Median from just objects near magerr cut
         
                mask_maglims[pix] = maglim

            logger.debug("%i (n=%i): maglim=%g"%(pix,num,mask_maglims[pix]))
            subpix = ugali.utils.skymap.subpixel(pix, self.nside_mask, self.nside_pixel)
            maglims = np.zeros(len(subpix)) + mask_maglims[pix] 
            out_pixels = np.append(out_pixels,subpix)
            out_maglims = np.append(out_maglims,maglims)
         
        # Remove empty pixels
        logger.info("Removing empty pixels")
        idx = np.nonzero(out_maglims > 0)[0]
        out_pixels  = out_pixels[idx]
        out_maglims = out_maglims[idx]
         
        # Remove pixels outside the footprint
        if self.footfile:
            logger.info("Checking footprint against %s"%self.footfile)
            lon,lat = pix2ang(self.nside_pixel,out_pixels)
            if self.config['coords']['coordsys'] == 'gal':
                ra,dec = gal2cel(lon,lat)
            else:    
                ra,dec = lon,lat
            footprint = inFootprint(self.footprint,ra,dec)
            idx = np.nonzero(footprint)[0]
            out_pixels = out_pixels[idx]
            out_maglims = out_maglims[idx]
         
        logger.info("MAGLIM = %.3f +/- %.3f"%(np.mean(out_maglims),np.std(out_maglims)))         
        return out_pixels,out_maglims

def inFootprint(footprint,ra,dec):
    """
    Check if set of ra,dec combinations are in footprint.
    Careful, input files must be in celestial coordinates.
    
    filename : Either healpix map or mangle polygon file
    ra,dec   : Celestial coordinates

    Returns:
    inside   : boolean array of coordinates in footprint
    """
    if footprint is None:
        return np.ones(len(ra),dtype=bool)
    
    try:
        if isinstance(footprint,str) and os.path.exists(footprint):
            filename = footprint
            #footprint = hp.read_map(filename,verbose=False)
            #footprint = fitsio.read(filename)['I'].ravel()
            footprint = read_map(filename)
        nside = hp.npix2nside(len(footprint))
        pix = ang2pix(nside,ra,dec)
        inside = (footprint[pix] > 0)
    except IOError:
        logger.warning("Failed to load healpix footprint; trying to use mangle...")
        inside = inMangle(filename,ra,dec)
    return inside

def inMangle(polyfile,ra,dec):
    coords = tempfile.NamedTemporaryFile(suffix='.txt',delete=False)
    logger.debug("Writing coordinates to %s"%coords.name)
    np.savetxt(coords, np.array( [ra,dec] ).T, fmt='%.6g' )
    coords.close()

    weights = tempfile.NamedTemporaryFile(suffix='.txt',delete=False)
    cmd = "polyid -W %s %s %s"%(polyfile,coords.name,weights.name)
    logger.debug(cmd)
    subprocess.call(cmd,shell=True)

    tmp = tempfile.NamedTemporaryFile(suffix='.txt',delete=False)
    cmd = """awk '{if($3==""){$3=0} print $1, $2, $3}' %s > %s"""%(weights.name,tmp.name)
    logger.debug(cmd)
    subprocess.call(cmd,shell=True)

    data = np.loadtxt(tmp.name,unpack=True,skiprows=1)[-1]
    for f in [coords,weights,tmp]:
        logger.debug("Removing %s"%f.name)
        os.remove(f.name)

    return data > 0

#def simple_maglims(config,dirname='simple',force=False):
#    """
#    Create simple, uniform magnitude limits based on nominal
#    survey depth.
#    """
#    DeprecationWarning("'simple_maglims' is deprecated")
#    filenames = config.getFilenames()
#    release = config['data']['release'].lower()
#    #band_1 = config['isochrone']['mag_1_field']
#    #band_2 = config['isochrone']['mag_2_field']
#    band_1 = config['catalog']['mag_1_field']
#    band_2 = config['catalog']['mag_2_field']
#    mask_1 = filenames['mask_1'].compressed()
#    mask_2 = filenames['mask_2'].compressed()
#    basedir,basename = os.path.split(config['mask']['dirname'])
#    if basename == dirname:
#        raise Exception("Input and output directory are the same.")
#    outdir = mkdir(os.path.join(basedir,dirname))
# 
#    for band, infiles in [(band_1,mask_1),(band_2,mask_2)]:
#        maglim = MAGLIMS[release][band]
#        for infile in infiles:
#            basename = os.path.basename(infile)
#            outfile = join(outdir,basename)
#            logger.debug('Reading %s...'%infile)
#            f = pyfits.open(infile)
#            f[1].data['MAGLIM'][:] = maglim
#            logger.debug('Writing %s...'%outfile)
#            f.writeto(outfile,clobber=True)

def split(config,dirname='split',force=False):
    """ Take a pre-existing maglim map and divide it into
    chunks consistent with the catalog pixels. """

    config = Config(config)
    filenames = config.getFilenames()
    #healpix = filenames['pix'].compressed()

    # Check that things are ok
    basedir,basename = os.path.split(config['mask']['dirname'])
    #if basename == dirname:
    #    msg = "Input and output directory are the same."
    #    raise Exception(msg)
    outdir = mkdir(os.path.join(basedir,dirname))
    
    nside_catalog = config['coords']['nside_catalog']
    nside_pixel = config['coords']['nside_pixel']

    release = config['data']['release'].lower()
    band1 = config['catalog']['mag_1_band']
    band2 = config['catalog']['mag_2_band']

    # Read the magnitude limits
    maglimdir = config['maglim']['dirname']

    maglimfile_1 = join(maglimdir,config['maglim']['filename_1'])
    logger.info("Reading %s..."%maglimfile_1)
    maglim1 = read_map(maglimfile_1)
    
    maglimfile_2 = join(maglimdir,config['maglim']['filename_2'])
    logger.info("Reading %s..."%maglimfile_2)
    maglim2 = read_map(maglimfile_2)

    # Read the footprint
    footfile = config['data']['footprint']
    logger.info("Reading %s..."%footfile)
    footprint = read_map(footfile)

    # Output mask names
    mask1 = os.path.basename(config['mask']['basename_1'])
    mask2 = os.path.basename(config['mask']['basename_2'])

    for band,maglim,base in [(band1,maglim1,mask1),(band2,maglim2,mask2)]:
        nside_maglim = hp.npix2nside(len(maglim))
        if nside_maglim != nside_pixel:
            msg = "Mask nside different from pixel nside"
            logger.warning(msg)
            #raise Exception(msg)

        pixels = np.nonzero(maglim>0)[0]
        superpix = superpixel(pixels,nside_maglim,nside_catalog)
        healpix = np.unique(superpix)
        for hpx in healpix:
            outfile = join(outdir,base)%hpx
            if os.path.exists(outfile) and not force:
                logger.warning("Found %s; skipping..."%outfile)
                continue

            pix = pixels[superpix == hpx]
            print(hpx, len(pix))

            logger.info('Writing %s...'%outfile)
            data = odict()
            data['PIXEL']=pix
            data['MAGLIM']=maglim[pix].astype('f4')
            data['FRACDET']=footprint[pix].astype('f4')
            ugali.utils.healpix.write_partial_map(outfile,data,nside_pixel)
                                                  
if __name__ == "__main__":
    from optparse import OptionParser
    usage = "Usage: %prog  [options] input"
    description = "python script"
    parser = OptionParser(usage=usage,description=description)
    (opts, args) = parser.parse_args()



#!/usr/bin/env python
"""
Generic python script.
"""
import os
from collections import OrderedDict as odict

from ugali.analysis.model import Model, Parameter
from ugali.utils.shell import mkdir, get_ugali_dir, get_iso_dir
from ugali.isochrone.parsec import PadovaIsochrone

class EmpiricalPadova(PadovaIsochrone):
    _prefix = 'iso'
    _basename = '%(prefix)s_a13.7_z0.00007.dat'
    _dirname =  os.path.join(get_iso_dir(),'{survey}','empirical')

    defaults = (PadovaIsochrone.defaults) + (
        ('dirname',_dirname,'Directory name for isochrone files'),
    )

class M92(EmpiricalPadova):
    """ Empirical isochrone derived from the M92 ridgeline dereddened
    and transformed to the DES system.
    """
    _params = odict([
        ('distance_modulus', Parameter(15.0, [10.0, 30.0]) ),
        ('age',              Parameter(13.7, [13.7, 13.7]) ),  # Gyr
        ('metallicity',      Parameter(7e-5,[7e-5,7e-5]) ),
    ])

    _prefix = 'm92'
    _basename = '%(prefix)s_a13.7_z0.00007.dat'

class DESDwarfs(EmpiricalPadova):
    """ Empirical isochrone derived from spectroscopic members of the
    DES dwarfs.
    """
    _params = odict([
        ('distance_modulus', Parameter(15.0, [10.0, 30.0]) ),
        ('age',              Parameter(12.5, [12.5, 12.5]) ),  # Gyr
        ('metallicity',      Parameter(1e-4, [1e-4,1e-4]) ),
    ])

    _prefix = 'dsph'
    _basename = '%(prefix)s_a12.5_z0.00010.dat'



#!/usr/bin/env python
"""
Module for various statistics utilities.
"""

import copy
from collections import OrderedDict as odict

import numpy as np
import numpy.lib.recfunctions as recfuncs
import scipy.special
import scipy.stats

# These should probably live in this file
from ugali.utils.bayesian_efficiency import bayesianInterval, binomialInterval
from ugali.utils import mlab

_alpha   = 0.32
_nbins   = 300
_npoints = 500

def mad_clip(data,mad=None,mad_lower=None,mad_upper=None):
    med = np.median(data)
    mad = np.median(np.fabs(med - data))
    if mad is not None:
        mad_lower = mad_upper = mad
    return 

def interval(best,lo=np.nan,hi=np.nan):
    """
    Pythonized interval for easy output to yaml
    """
    return [float(best),[float(lo),float(hi)]]

def mean_interval(data, alpha=_alpha):
    """
    Interval assuming gaussian posterior.
    """
    mean = np.mean(data)
    sigma = np.std(data)
    scale = scipy.stats.norm.ppf(1-alpha/2.)
    return interval(mean,mean-scale*sigma,mean+scale*sigma)

def median_interval(data, alpha=_alpha):
    """
    Median with bayesian credible interval from percentiles.
    """
    q = [100*alpha/2., 50, 100*(1-alpha/2.)]
    lo,med,hi = np.percentile(data,q)
    return interval(med,lo,hi)
    
def peak(data, bins=_nbins):
    """
    Bin the distribution and find the mode

    Parameters:
    -----------
    data  : The 1d data sample
    bins  : Number of bins

    Returns
    -------
    peak : peak of the kde
    """
    num,edges = np.histogram(data,bins=bins)
    centers = (edges[1:]+edges[:-1])/2.
    return centers[np.argmax(num)]

def kde_peak(data, npoints=_npoints, clip=5.0):
    """
    Identify peak using Gaussian kernel density estimator.

    Parameters:
    -----------
    data    : The 1d data sample
    npoints : The number of kde points to evaluate
    clip    : NMAD to clip

    Returns
    -------
    peak : peak of the kde
    """
    return kde(data,npoints,clip)[0]

def kde(data, npoints=_npoints, clip=5.0):
    """
    Identify peak using Gaussian kernel density estimator.
    
    Parameters:
    -----------
    data    : The 1d data sample
    npoints : The number of kde points to evaluate
    clip    : NMAD to clip

    Returns
    -------
    peak : peak of the kde
    """

    # Clipping of severe outliers to concentrate more KDE samples
    # in the parameter range of interest
    mad = np.median(np.fabs(np.median(data) - data))
    if clip > 0:
        cut  = (data > np.median(data) - clip * mad)
        cut &= (data < np.median(data) + clip * mad)
        x = data[cut]
    else:
        x = data
    kde = scipy.stats.gaussian_kde(x)
    # No penalty for using a finer sampling for KDE evaluation
    # except computation time
    values = np.linspace(np.min(x), np.max(x), npoints)
    kde_values = kde.evaluate(values)
    peak = values[np.argmax(kde_values)]
    return peak, kde.evaluate(peak)

def peak_interval(data, alpha=_alpha, npoints=_npoints):
    """Identify minimum interval containing the peak of the posterior as
    determined by a Gaussian kernel density estimator.

    Parameters
    ----------
    data   : the 1d data sample
    alpha  : the confidence interval
    npoints: number of kde points to evaluate

    Returns
    -------
    interval : the minimum interval containing the peak
    """
    peak = kde_peak(data,npoints)
    x = np.sort(data.flat); n = len(x)
    # The number of entries in the interval
    window = int(np.rint((1.0-alpha)*n))
    # The start, stop, and width of all possible intervals
    starts = x[:n-window]; ends = x[window:]
    widths = ends - starts
    # Just the intervals containing the peak
    select = (peak >= starts) & (peak <= ends)
    widths = widths[select]
    if len(widths) == 0:
        raise ValueError('Too few elements for interval calculation')
    min_idx = np.argmin(widths)
    lo = starts[select][min_idx]
    hi = ends[select][min_idx]
    return interval(peak,lo,hi)

def min_interval(data, alpha=_alpha):
    """Minimum interval containing 1-alpha of the posterior.
    Note: interval is *not* required to contain the peak of the
    posterior.

    Parameters
    ----------
    data   : the 1d data sample
    alpha  : the confidence interval

    Returns
    -------
    interval : the minimum interval
    """
    x = np.sort(data.flat); n = len(x)
    # The number of entries in the interval
    window = int(np.rint((1.0-alpha)*n))
    # The start, stop, and width of all possible intervals
    starts = x[:n-window]; ends = x[window:]
    widths = ends - starts
    if len(widths) == 0:
        raise ValueError('Too few elements for interval calculation')
    min_idx = np.argmin(widths)
    lo = starts[min_idx]
    hi = ends[min_idx]
    center = (hi+lo)/2.
    return interval(center,lo,hi)

def norm_cdf(x):
    """Faster than scipy.stats.norm.cdf
    https://en.wikipedia.org.wiki/Normal_distribution
    """
    return 0.5*(1 + scipy.special.erf(x/np.sqrt(2)))

def random_pdf(value,pdf,size=None):
    if size is None: size = 1.0
    cdf = np.cumsum(pdf)
    cdf /= cdf[-1]
    fn = scipy.interpolate.interp1d(cdf, list(range(0, len(cdf))))
    index = np.rint(fn(np.random.uniform(size=size))).astype(int)
    return value[index]

def sky(lon=None,lat=None,size=1):
    """
    Outputs uniform points on sphere from:
    [0 < lon < 360] & [-90 < lat < 90]
    """
    if lon is None:
        umin,umax = 0,1
    else:
        lon = np.asarray(lon)
        lon = np.radians(lon + 360.*(lon<0))
        if   lon.size==1: umin=umax=lon/(2*np.pi)
        elif lon.size==2: umin,umax=lon/(2*np.pi)
        else: raise Exception('...')
        
    if lat is None:
        vmin,vmax = -1,1
    else:
        lat = np.asarray(lat)
        lat = np.radians(90 - lat)
        if   lat.size==1: vmin=vmax=np.cos(lat)
        elif lat.size==2: vmin,vmax=np.cos(lat)
        else: raise Exception('...')

    phi = 2*np.pi*np.random.uniform(umin,umax,size=size)
    theta = np.arcsin(np.random.uniform(vmin,vmax,size=size))
    return np.degrees(phi),np.degrees(theta)


class Samples(np.recarray):
    """
    Wrapper class for recarray to deal with MCMC samples.
    
    A nice summary of various bayesian credible intervals can be found here:
    http://www.sumsar.net/blog/2014/10/probable-points-and-credible-intervals-part-one/
    """
    _alpha   = 0.10
    _nbins   = 300
    _npoints = 250

    def __new__(cls, input, names=None):
        # Load the array from file
        if not isinstance(input,np.ndarray):
            obj = np.load(input).view(cls)
        else:
            obj = np.asarray(input).view(cls)
            
        # (re)set the column names
        if names is not None:
            if obj.dtype.names is None:
                obj = np.rec.fromarrays(obj,names=names).view(cls)
            else:
                obj.dtype.names = names

        return obj

    def __array_wrap__(self, out_arr, context=None):
        return np.ndarray.__array_wrap__(self,out_arr,context)

    @property
    def names(self):
        return self.dtype.names

    @property
    def ndarray(self):
        # atleast_2d is for 
        if len(self.dtype) == 1:
            return np.expand_dims(self.view((float,len(self.dtype))),1)
        else:
            return self.view((float,len(self.dtype)))

    def supplement(self,coordsys='gal'):
        """ Add some supplemental columns """
        from ugali.utils.projector import gal2cel, gal2cel_angle
        from ugali.utils.projector import cel2gal, cel2gal_angle

        coordsys = coordsys.lower()
        kwargs = dict(usemask=False, asrecarray=True)
        out = copy.deepcopy(self)

        if ('lon' in out.names) and ('lat' in out.names):
            # Ignore entries that are all zero
            zeros = np.all(self.ndarray==0,axis=1)

            if coordsys == 'gal':
                ra,dec = gal2cel(out.lon,out.lat)
                glon,glat = out.lon,out.lat
            else:
                ra,dec = out.lon,out.lat
                glon,glat = cel2gal(out.lon,out.lat)

            ra[zeros] = 0; dec[zeros] = 0
            glon[zeros] = 0; glat[zeros] = 0

            names = ['ra','dec','glon','glat']
            arrs = [ra,dec,glon,glat]
            out = mlab.rec_append_fields(out,names,arrs).view(Samples)
            #out = recfuncs.append_fields(out,names,arrs,**kwargs).view(Samples)

            if 'position_angle' in out.names:
                if coordsys == 'gal':
                    pa_gal = out.position_angle
                    pa_cel = gal2cel_angle(out.lon,out.lat,out.position_angle)
                    pa_cel = pa_cel - 180.*(pa_cel > 180.)
                else:
                    pa_gal = cel2gal_angle(out.lon,out.lat,out.position_angle)
                    pa_cel = out.position_angle
                    pa_gal = pa_gal - 180.*(pa_gal > 180.)
                    
                pa_gal[zeros] = 0; pa_cel[zeros] = 0
                names = ['position_angle_gal','position_angle_cel']
                arrs = [pa_gal,pa_cel]
                out = recfuncs.append_fields(out,names,arrs,**kwargs).view(Samples)
        
        return out

    def get(self, names=None, burn=None, clip=None):
        if names is None: names = list(self.dtype.names)
        names = np.array(names,ndmin=1)

        missing = names[~np.in1d(names,self.dtype.names)]
        if len(missing):
            msg = "field(s) named %s not found"%(missing)
            raise ValueError(msg)
        #idx = np.where(np.in1d(self.dtype.names,names))[0]
        idx = np.array([self.dtype.names.index(n) for n in names])

        # Remove zero entries
        zsel = ~np.all(self.ndarray==0,axis=1)
        # Remove burn entries
        bsel = np.zeros(len(self),dtype=bool)
        bsel[slice(burn,None)] = 1

        data = self.ndarray[:,idx][bsel&zsel]
        if clip is not None:
            from astropy.stats import sigma_clip
            mask = sigma_clip(data,sigma=clip,copy=False,axis=0).mask
            data = data[np.where(~mask.any(axis=1))]

        return data


    @classmethod
    def _interval(cls,best,lo,hi):
        """
        Pythonized interval for easy output to yaml
        """
        #return ugali.utils.stats.interval(best,lo,hi)
        return interval(best,lo,hi)

    def mean(self, name, **kwargs):
        """
        Mean of the distribution.
        """
        return np.mean(self.get(name,**kwargs))

    def mean_interval(self, name, alpha=_alpha, **kwargs):
        """
        Interval assuming gaussian posterior.
        """
        data = self.get(name,**kwargs)
        #return ugali.utils.stats.mean_interval(data,alpha)
        return mean_interval(data,alpha)

    def median(self, name, **kwargs):
        """
        Median of the distribution.
        """
        data = self.get(name,**kwargs)
        return np.percentile(data,[50])

    def median_interval(self, name, alpha=_alpha, **kwargs):
        """
        Median including bayesian credible interval.
        """
        data = self.get(name,**kwargs)
        return median_interval(data,alpha)

    def peak(self, name, bins=_nbins, **kwargs):
        data = self.get(name,**kwargs)
        return peak(data,bins=bins)

    def kde_peak(self, name, npoints=_npoints, **kwargs):
        """ 
        Calculate peak of kernel density estimator
        """
        data = self.get(name,**kwargs)
        return kde_peak(data,npoints)

    def kde(self, name, npoints=_npoints, **kwargs):
        """ 
        Calculate kernel density estimator for parameter
        """
        data = self.get(name,**kwargs)
        return kde(data,npoints)

    def peak_interval(self, name, alpha=_alpha, npoints=_npoints, **kwargs):
        """ 
        Calculate peak interval for parameter.
        """
        data = self.get(name, **kwargs)
        return peak_interval(data,alpha,npoints)

    def min_interval(self,name, alpha=_alpha, **kwargs):
        """ 
        Calculate minimum interval for parameter.
        """
        data = self.get(name, **kwargs)
        return min_interval(data,alpha)

    def results(self, names=None, alpha=_alpha, mode='peak', **kwargs):
        """
        Calculate the results for a set of parameters.
        """
        if names is None: names = self.names
        ret = odict()
        for n in names:
            ret[n] = getattr(self,'%s_interval'%mode)(n, **kwargs)
        return ret


if __name__ == "__main__":
    import argparse
    description = "python script"
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('args',nargs=argparse.REMAINDER)
    opts = parser.parse_args(); args = opts.args

    import pylab as plt
    ax=plt.subplot(221,projection='aitoff')
    ax.grid(True)
    lon,lat = sky(size=1e3)
    lon,lat= np.radians([lon-360.*(lon>180),lat])
    ax.scatter(lon,lat,marker='.',s=2)

    ax=plt.subplot(222,projection='aitoff')
    ax.grid(True)
    lon,lat = sky(size=1e3,lat=[30,45])
    lon,lat= np.radians([lon-360.*(lon>180),lat])
    ax.scatter(lon,lat,marker='.',s=2)

    ax=plt.subplot(223,projection='aitoff')
    ax.grid(True)
    lon,lat = sky(size=1e3,lon=[30,45])
    lon,lat= np.radians([lon-360.*(lon>180),lat])
    ax.scatter(lon,lat,marker='.',s=2)

    ax=plt.subplot(224,projection='aitoff')
    ax.grid(True)
    lon,lat = sky(size=1e3,lon=[0,45],lat=[30,45])
    lon,lat= np.radians([lon-360.*(lon>180),lat])
    ax.scatter(lon,lat,marker='.',s=2)



"""
Classes which manage object catalogs live here.
"""
import numpy as np
import fitsio
import copy

import ugali.utils.projector

from ugali.utils.config import Config
from ugali.utils.projector import gal2cel,cel2gal
from ugali.utils.healpix import ang2pix,superpixel
from ugali.utils.logger import logger
from ugali.utils.fileio import load_infiles
from ugali.utils import mlab

class Catalog:

    def __init__(self, config, roi=None, data=None, filenames=None):
        """
        Class to store information about detected objects. This class
        augments the raw data array with several aliases and derived
        quantities.

        Parameters:
        -----------
        config    : Configuration object
        roi       : Region of Interest to load catalog data for
        data      : Data array object
        filenames : FITS filenames to read catalog from

        Returns:
        --------
        catalog   : The Catalog object
        """
        self.config = Config(config)

        if data is None:
            self._parse(roi,filenames)
        else:
            self.data = data

        self._defineVariables()

    def __add__(self, other):
        return mergeCatalogs([self,other])

    def __len__(self):
        return len(self.objid)

    def __eq__(self, other):
        return bool((self.data == other.data).all())
        
    def applyCut(self, cut):
        """
        Return a new catalog which is a subset of objects selected
        using the input cut array.

        NOTE: This is really a *selection* (i.e., objects are retained if the value of 'cut' is True)
        """
        return Catalog(self.config, data=self.data[cut])

    def bootstrap(self, mc_bit=0x10, seed=None):
        """
        Return a random catalog by boostrapping the colors of the objects in the current catalog.
        """
        if seed is not None: np.random.seed(seed)
        data = copy.deepcopy(self.data)
        idx = np.random.randint(0,len(data),len(data))
        data[self.config['catalog']['mag_1_field']][:] = self.mag_1[idx]
        data[self.config['catalog']['mag_err_1_field']][:] = self.mag_err_1[idx]
        data[self.config['catalog']['mag_2_field']][:] = self.mag_2[idx]
        data[self.config['catalog']['mag_err_2_field']][:] = self.mag_err_2[idx]
        data[self.config['catalog']['mc_source_id_field']][:] |= mc_bit
        return Catalog(self.config, data=data)

    def project(self, projector = None):
        """
        Project coordinates on sphere to image plane using Projector class.
        """
        msg = "'%s.project': ADW 2018-05-05"%self.__class__.__name__
        DeprecationWarning(msg)
        if projector is None:
            try:
                self.projector = ugali.utils.projector.Projector(self.config['coords']['reference'][0],
                                                                 self.config['coords']['reference'][1])
            except KeyError:
                logger.warning('Projection reference point is median (lon, lat) of catalog objects')
                self.projector = ugali.utils.projector.Projector(np.median(self.lon), np.median(self.lat))
        else:
            self.projector = projector

        self.x, self.y = self.projector.sphereToImage(self.lon, self.lat)

    def spatialBin(self, roi):
        """
        Calculate indices of ROI pixels corresponding to object locations.
        """
        if hasattr(self,'pixel_roi_index') and hasattr(self,'pixel'): 
            logger.warning('Catalog alread spatially binned')
            return

        # ADW: Not safe to set index = -1 (since it will access last entry); 
        # np.inf would be better...
        self.pixel = ang2pix(self.config['coords']['nside_pixel'],self.lon,self.lat)
        self.pixel_roi_index = roi.indexROI(self.lon,self.lat)

        logger.info("Found %i objects outside ROI"%(self.pixel_roi_index < 0).sum())

    def write(self, outfile, clobber=True, **kwargs):
        """
        Write the current object catalog to FITS file.

        Parameters:
        -----------
        filename : the FITS file to write.
        clobber  : remove existing file
        kwargs   : passed to fitsio.write

        Returns:
        --------
        None
        """
        fitsio.write(outfile,self.data,clobber=True,**kwargs)

    def _parse(self, roi=None, filenames=None):
        """        
        Parse catalog FITS files into recarray.

        Parameters:
        -----------
        roi : The region of interest; if 'roi=None', read all catalog files

        Returns:
        --------
        None
        """
        if (roi is not None) and (filenames is not None):
            msg = "Cannot take both roi and filenames"
            raise Exception(msg)

        if roi is not None:
            pixels = roi.getCatalogPixels()
            filenames = self.config.getFilenames()['catalog'][pixels]
        elif filenames is None:
            filenames = self.config.getFilenames()['catalog'].compressed()
        else:
            filenames = np.atleast_1d(filenames)

        if len(filenames) == 0:
            msg = "No catalog files found."
            raise Exception(msg)

        # Load the data
        self.data = load_infiles(filenames)

        # Apply a selection cut
        self._applySelection()

        # Cast data to recarray (historical reasons)
        self.data = self.data.view(np.recarray)

    def _applySelection(self,selection=None):
        # ADW: This is a hack (eval is unsafe!)
        if selection is None:
            selection = self.config['catalog'].get('selection')

        if not selection: 
            return
        elif 'self.data' not in selection:
            msg = "Selection does not contain 'data'"
            raise Exception(msg)
        else:
            logger.info('Evaluating selection: \n"%s"'%selection)
            sel = eval(selection)
            self.data = self.data[sel]
        
    def _defineVariables(self):
        """
        Helper funtion to define pertinent variables from catalog data.

        ADW (20170627): This has largely been replaced by properties.
        """
        logger.info('Catalog contains %i objects'%(len(self.data)))

        mc_source_id_field = self.config['catalog']['mc_source_id_field']
        if mc_source_id_field is not None:
            if mc_source_id_field not in self.data.dtype.names:
                array = np.zeros(len(self.data),dtype='>i8') # FITS byte-order convention
                self.data = mlab.rec_append_fields(self.data,
                                                   names=mc_source_id_field,
                                                   arrs=array)
            logger.info('Found %i simulated objects'%(np.sum(self.mc_source_id>0)))

    # Use properties to avoid duplicating the data
    @property
    def objid(self): return self.data[self.config['catalog']['objid_field']]
    @property
    def lon(self): return self.data[self.config['catalog']['lon_field']]
    @property 
    def lat(self): return self.data[self.config['catalog']['lat_field']]
    @property
    def coordsys(self): 
        return self.config['coords']['coordsys'].lower()

    @property
    def mag_1(self): return self.data[self.config['catalog']['mag_1_field']]
    @property
    def mag_err_1(self):
        return self.data[self.config['catalog']['mag_err_1_field']]
    @property
    def mag_2(self):
        return self.data[self.config['catalog']['mag_2_field']]
    @property
    def mag_err_2(self):
        return self.data[self.config['catalog']['mag_err_2_field']]
    @property 
    def mag(self):
        if self.config['catalog']['band_1_detection']: return self.mag_1
        else: return self.mag_2
    @property
    def mag_err(self):
        if self.config['catalog']['band_1_detection']: return self.mag_err_1
        else: return self.mag_err_2
    @property
    def color(self): return self.mag_1 - self.mag_2
    @property
    def color_err(self): return np.sqrt(self.mag_err_1**2 + self.mag_err_2**2)
    @property
    def mc_source_id(self):
        return self.data[self.config['catalog']['mc_source_id_field']]

    # This assumes Galactic coordinates
    @property
    def ra_dec(self): 
        if self.coordsys == 'cel': return self.lon, self.lat
        else:                      return gal2cel(self.lon,self.lat)
    @property
    def ra(self): return self.ra_dec[0]
    @property
    def dec(self): return self.ra_dec[1]

    @property
    def glon_glat(self): 
        if self.coordsys == 'gal': return self.lon,self.lat
        else:                      return cel2gal(self.lon,self.lat)
    @property
    def glon(self): return self.glon_glat[0]
    @property
    def glat(self): return self.glon_glat[1]

############################################################

def mergeCatalogs(catalog_list):
    """
    Merge a list of Catalogs.

    Parameters:
    -----------
    catalog_list : List of Catalog objects.

    Returns:
    --------
    catalog      : Combined Catalog object 
    """
    # Check the columns
    for c in catalog_list:
        if c.data.dtype.names != catalog_list[0].data.dtype.names:
            msg = "Catalog data columns not the same."
            raise Exception(msg)
    data = np.concatenate([c.data for c in catalog_list])
    config = catalog_list[0].config
    return Catalog(config,data=data)
